# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:49 PM

"""All Protobuf defnitions."""

from open_gopro.proto.turbo_transfer_pb2 import RequestSetTurboActive
from open_gopro.proto.response_generic_pb2 import ResponseGeneric
from open_gopro.proto.request_get_preset_status_pb2 import RequestGetPresetStatus
from open_gopro.proto.preset_status_pb2 import NotifyPresetStatus
from open_gopro.proto.set_camera_control_status_pb2 import RequestSetCameraControlStatus
from open_gopro.proto.live_streaming_pb2 import (
    RequestSetLiveStreamMode,
    RequestGetLiveStreamStatus,
    NotifyLiveStreamStatus,
)
from open_gopro.proto.network_management_pb2 import (
    RequestConnectNew,
    ResponseConnectNew,
    NotifProvisioningState,
    RequestStartScan,
    ResponseStartScanning,
    RequestGetApEntries,
    ResponseGetApEntries,
    NotifStartScanning,
    RequestReleaseNetwork,
)



# api.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""Implementation of Open GoPro API version 2.0"""

from typing import Final

from open_gopro.interface import GoProInterface
from .ble_commands import BleCommands, BleSettings, BleStatuses, BleAsyncResponses
from .wifi_commands import WifiCommands, WifiSettings


class Api:
    """Implementation of Open GoPro API version 2.0"""

    version: Final = "2.0"

    def __init__(self, communicator: GoProInterface) -> None:
        """Constructor

        Args:
            communicator (GoProInterface): used to communicate via BLE and Wifi
        """
        self._communicator = communicator
        self.ble_command = BleCommands(communicator)
        self.ble_setting = BleSettings(communicator)
        self.ble_status = BleStatuses(communicator)
        BleAsyncResponses.add_parsers()
        self.wifi_command = WifiCommands(communicator)
        self.wifi_setting = WifiSettings(communicator)



# constants.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:44 PM

"""Constant numbers shared across the GoPro module. These do not change across Open GoPro Versions"""

from __future__ import annotations
from enum import IntEnum, Enum, IntFlag, EnumMeta
from dataclasses import dataclass
from typing import Protocol, Union, Iterator, TypeVar, Any, no_type_check, Final

import construct

from open_gopro.ble import BleUUID, UUIDs

T = TypeVar("T")


##############################################################################################################
#####################   Custom Enum for this Project  ########################################################
##############################################################################################################


class ProtobufDescriptor(Protocol):
    """Protocol definition for Protobuf enum descriptor used to generate GoPro enums from protobufs"""

    @property
    def name(self) -> str:
        """Human readable name of protobuf enum

        # noqa: DAR202

        Returns:
            str: enum name
        """

    @property
    def values_by_name(self) -> dict:
        """Get the enum values by name

        # noqa: DAR202

        Returns:
            dict: Dict of enum values mapped by name
        """

    @property
    def values_by_number(self) -> dict:
        """Get the enum values by number

        # noqa: DAR202

        Returns:
            dict: dict of enum numbers mapped by numberf
        """


class GoProEnumMeta(EnumMeta):
    """Modify enum metaclass to build GoPro specific enums"""

    _is_proto = False
    _iter_skip_names = ("NOT_APPLICABLE", "DESCRIPTOR")

    @no_type_check
    def __new__(cls, name, bases, classdict, **kwargs) -> GoProEnumMeta:  # noqa
        is_proto = "__is_proto__" in classdict
        classdict["_ignore_"] = "__is_proto__"
        classdict["__doc__"] = ""  # Don't use useless "An enumeration" docstring
        e = super().__new__(cls, name, bases, classdict, **kwargs)
        setattr(e, "_is_proto", is_proto)
        return e

    @no_type_check
    def __contains__(cls: type[Any], obj: object) -> bool:
        if isinstance(obj, Enum):
            return super().__contains__(obj)
        if isinstance(obj, int):
            return obj in [x.value for x in cls._member_map_.values()]
        if isinstance(obj, str):
            return obj.lower() in [x.name.lower() for x in cls._member_map_.values()]
        raise TypeError(
            f"unsupported operand type(s) for 'in': {type(obj).__qualname__} and {cls.__class__.__qualname__}"
        )

    def __iter__(cls: type[T]) -> Iterator[T]:
        """Do not return enum values whose name is in the _iter_skip_names list

        Returns:
            Iterator[T]: enum iterator
        """
        return iter([x[1] for x in cls._member_map_.items() if x[0] not in GoProEnumMeta._iter_skip_names])  # type: ignore


class GoProFlagEnum(IntFlag, metaclass=GoProEnumMeta):
    """GoPro specific enum to be used for all settings, statuses, and parameters

    The names NOT_APPLICABLE and DESCRIPTOR are special as they will not be returned as part of the enum iterator
    """

    def __eq__(self, other: object) -> bool:
        if type(self)._is_proto:
            if isinstance(other, int):
                return self.value == other
            if isinstance(other, str):
                return self.name == other
            if isinstance(other, Enum):
                return self.value == other.value
            raise TypeError(f"Unexpected case: proto enum can only be str or int, not {type(other)}")
        return super().__eq__(other)

    def __hash__(self) -> Any:
        return hash(self.name or "" + str(self.value))


class GoProEnum(IntEnum, metaclass=GoProEnumMeta):
    """GoPro specific enum to be used for all settings, statuses, and parameters

    The names NOT_APPLICABLE and DESCRIPTOR are special as they will not be returned as part of the enum iterator
    """

    def __eq__(self, other: object) -> bool:
        if type(self)._is_proto:
            if isinstance(other, int):
                return self.value == other
            if isinstance(other, str):
                return self.name == other
            if isinstance(other, Enum):
                return self.value == other.value
            raise TypeError(f"Unexpected case: proto enum can only be str or int, not {type(other)}")
        return super().__eq__(other)

    def __hash__(self) -> Any:
        return hash(self.name + str(self.value))


def enum_factory(proto_enum: ProtobufDescriptor) -> type[GoProEnum]:
    """Dynamically build a GoProEnum from a protobuf enum

    Args:
        proto_enum (ProtobufDescriptor): input protobuf enum descriptor

    Returns:
        GoProEnum: generated GoProEnum
    """
    return GoProEnum(  # type: ignore # pylint: disable=too-many-function-args
        proto_enum.name,  # type: ignore
        {
            **dict(
                zip(
                    proto_enum.values_by_name.keys(),
                    proto_enum.values_by_number.keys(),
                )
            ),
            "__is_proto__": True,
        },
    )


##############################################################################################################
#####################  End Enum Definition ###################################################################
##############################################################################################################

GOPRO_BASE_UUID: Final = "b5f9{}-aa8d-11e3-9046-0002a5d5c51b"


@dataclass(frozen=True)
class GoProUUIDs(UUIDs):
    """GoPro Proprietary BleUUID's."""

    # GoPro Wifi Access Point Service
    S_WIFI_ACCESS_POINT = BleUUID("Wifi Access Point Service", hex=GOPRO_BASE_UUID.format("0001"))
    WAP_SSID = BleUUID("Wifi AP SSID", hex=GOPRO_BASE_UUID.format("0002"))
    WAP_PASSWORD = BleUUID("Wifi AP Password", hex=GOPRO_BASE_UUID.format("0003"))
    WAP_POWER = BleUUID("Wifi Power", hex=GOPRO_BASE_UUID.format("0004"))
    WAP_STATE = BleUUID("Wifi State", hex=GOPRO_BASE_UUID.format("0005"))
    WAP_CSI_PASSWORD = BleUUID("CSI Password", hex=GOPRO_BASE_UUID.format("0006"))

    # GoPro Control & Query Service
    S_CONTROL_QUERY = BleUUID("Control and Query Service", hex="0000fea6-0000-1000-8000-00805f9b34fb")
    CQ_COMMAND = BleUUID("Command", hex=GOPRO_BASE_UUID.format("0072"))
    CQ_COMMAND_RESP = BleUUID("Command Response", hex=GOPRO_BASE_UUID.format("0073"))
    CQ_SETTINGS = BleUUID("Settings", hex=GOPRO_BASE_UUID.format("0074"))
    CQ_SETTINGS_RESP = BleUUID("Settings Response", hex=GOPRO_BASE_UUID.format("0075"))
    CQ_QUERY = BleUUID("Query", hex=GOPRO_BASE_UUID.format("0076"))
    CQ_QUERY_RESP = BleUUID("Query Response", hex=GOPRO_BASE_UUID.format("0077"))
    CQ_SENSOR = BleUUID("Sensor", hex=GOPRO_BASE_UUID.format("0078"))
    CQ_SENSOR_RESP = BleUUID("Sensor Response", hex=GOPRO_BASE_UUID.format("0079"))

    # GoPro Camera Management Service
    S_CAMERA_MANAGEMENT = BleUUID("Camera Management Service", hex=GOPRO_BASE_UUID.format("0090"))
    CM_NET_MGMT_COMM = BleUUID("Camera Management", hex=GOPRO_BASE_UUID.format("0091"))
    CN_NET_MGMT_RESP = BleUUID("Camera Management Response", hex=GOPRO_BASE_UUID.format("0092"))

    # Unknown
    S_INTERNAL = BleUUID("Unknown Service", hex=GOPRO_BASE_UUID.format("0080"))
    INTERNAL_81 = BleUUID("Internal 81", hex=GOPRO_BASE_UUID.format("0081"))
    INTERNAL_82 = BleUUID("Internal 82", hex=GOPRO_BASE_UUID.format("0082"))
    INTERNAL_83 = BleUUID("Internal 83", hex=GOPRO_BASE_UUID.format("0083"))
    INTERNAL_84 = BleUUID("Internal 84", hex=GOPRO_BASE_UUID.format("0084"))


class ErrorCode(GoProEnum):
    """Status Codes."""

    SUCCESS = 0
    ERROR = 1
    INVALID_PARAM = 2
    UNKNOWN = -1


class CmdId(GoProEnum):
    """Command ID's that are written to GoProUUIDs.CQ_COMMAND."""

    SET_SHUTTER = 0x01
    POWER_DOWN = 0x04
    SLEEP = 0x05
    SET_PAIRING_COMPLETE = 0x03
    SET_DATE_TIME = 0x0D
    GET_DATE_TIME = 0x0E
    SET_DATE_TIME_DST = 0x0F
    GET_DATE_TIME_DST = 0x10
    GET_CAMERA_SETTINGS = 0x12
    GET_CAMERA_STATUSES = 0x13
    GET_CAMERA_CAPABILITIES = 0x32
    SET_WIFI = 0x17
    TAG_HILIGHT = 0x18
    GET_SETTINGS_JSON = 0x3B
    GET_HW_INFO = 0x3C
    LOAD_PRESET_GROUP = 0x3E
    LOAD_PRESET = 0x40
    SET_THIRD_PARTY_CLIENT_INFO = 0x50
    GET_THIRD_PARTY_API_VERSION = 0x51
    REGISTER_ALL_SETTINGS = 0x52
    REGISTER_ALL_STATUSES = 0x53
    UNREGISTER_ALL_SETTINGS = 0x72
    UNREGISTER_ALL_STATUSES = 0x73
    REGISTER_ALL_CAPABILITIES = 0x62
    UNREGISTER_ALL_CAPABILITIES = 0x82


class ActionId(GoProEnum):
    """Action ID's that identify a protobuf command."""

    SCAN_WIFI_NETWORKS = 0x02
    GET_AP_ENTRIES = 0x03
    REQUEST_WIFI_CONNECT = 0x05
    NOTIF_START_SCAN = 0x0B
    NOTIF_PROVIS_STATE = 0x0C
    SET_CAMERA_CONTROL = 0x69
    SET_TURBO_MODE = 0x6B
    GET_PRESET_STATUS = 0x72
    GET_LIVESTREAM_STATUS = 0x74
    RELEASE_NETWORK = 0x78
    SET_LIVESTREAM_MODE = 0x79
    SCAN_WIFI_NETWORKS_RSP = 0x82
    GET_AP_ENTRIES_RSP = 0x83
    REQUEST_WIFI_CONNECT_RSP = 0x85
    SET_CAMERA_CONTROL_RSP = 0xE9
    SET_TURBO_MODE_RSP = 0xEB
    GET_PRESET_STATUS_RSP = 0xF2
    PRESET_MODIFIED_NOTIFICATION = 0xF3
    LIVESTREAM_STATUS_RSP = 0xF4
    LIVESTREAM_STATUS_NOTIF = 0xF5
    RELEASE_NETWORK_RSP = 0xF8
    SET_LIVESTREAM_MODE_RSP = 0xF9
    INTERNAL_FF = 0xFF


class FeatureId(GoProEnum):
    """ID's that group protobuf commands"""

    NETWORK_MANAGEMENT = 0x02
    COMMAND = 0xF1
    SETTING = 0xF3
    QUERY = 0xF5


class SettingId(GoProEnum):
    """Setting ID's that identify settings and are written to GoProUUIDs.CQ_SETTINGS."""

    RESOLUTION = 2
    FPS = 3
    INTERNAL_5 = 5
    INTERNAL_6 = 6
    INTERNAL_13 = 13
    INTERNAL_19 = 19
    INTERNAL_24 = 24
    INTERNAL_30 = 30
    INTERNAL_31 = 31
    INTERNAL_32 = 32
    INTERNAL_37 = 37
    INTERNAL_41 = 41
    INTERNAL_42 = 42
    INTERNAL_43 = 43
    INTERNAL_44 = 44
    INTERNAL_45 = 45
    INTERNAL_47 = 47
    INTERNAL_48 = 48
    INTERNAL_54 = 54
    AUTO_OFF = 59
    INTERNAL_60 = 60
    INTERNAL_61 = 61
    INTERNAL_62 = 62
    INTERNAL_64 = 64
    INTERNAL_65 = 65
    INTERNAL_66 = 66
    INTERNAL_67 = 67
    INTERNAL_68 = 68
    INTERNAL_69 = 69
    INTERNAL_75 = 75
    INTERNAL_76 = 76
    INTERNAL_79 = 79
    INTERNAL_83 = 83
    INTERNAL_84 = 84
    INTERNAL_85 = 85
    INTERNAL_86 = 86
    INTERNAL_87 = 87
    INTERNAL_88 = 88
    LED = 91
    INTERNAL_96 = 96
    INTERNAL_102 = 102
    INTERNAL_103 = 103
    INTERNAL_104 = 104
    INTERNAL_105 = 105
    INTERNAL_106 = 106
    INTERNAL_111 = 111
    INTERNAL_112 = 112
    INTERNAL_114 = 114
    INTERNAL_115 = 115
    INTERNAL_116 = 116
    INTERNAL_117 = 117
    INTERNAL_118 = 118
    VIDEO_FOV = 121
    PHOTO_FOV = 122
    MULTI_SHOT_FOV = 123
    INTERNAL_124 = 124
    INTERNAL_125 = 125
    INTERNAL_126 = 126
    MEDIA_FORMAT = 128
    INTERNAL_129 = 129
    INTERNAL_130 = 130
    INTERNAL_131 = 131
    INTERNAL_132 = 132
    INTERNAL_133 = 133
    ANTI_FLICKER = 134
    HYPERSMOOTH = 135
    INTERNAL_139 = 139
    INTERNAL_142 = 142
    INTERNAL_144 = 144
    INTERNAL_145 = 145
    INTERNAL_146 = 146
    INTERNAL_147 = 147
    INTERNAL_148 = 148
    INTERNAL_149 = 149
    VIDEO_HORIZON_LEVELING = 150
    PHOTO_HORIZON_LEVELING = 151
    INTERNAL_153 = 153
    INTERNAL_154 = 154
    INTERNAL_155 = 155
    INTERNAL_156 = 156
    INTERNAL_157 = 157
    INTERNAL_158 = 158
    INTERNAL_159 = 159
    INTERNAL_160 = 160
    INTERNAL_161 = 161
    MAX_LENS_MOD = 162
    INTERNAL_163 = 163
    INTERNAL_164 = 164
    INTERNAL_165 = 165
    INTERNAL_166 = 166
    INTERNAL_167 = 167
    INTERNAL_168 = 168
    INTERNAL_169 = 169
    VIDEO_PERFORMANCE_MODE = 173
    INTERNAL_174 = 174
    CAMERA_UX_MODE = 175
    VIDEO_EASY_MODE = 176
    PHOTO_EASY_MODE = 177
    WIFI_BAND = 178
    STAR_TRAIL_LENGTH = 179
    SYSTEM_VIDEO_MODE = 180
    PROTOBUF_SETTING = 0xF3
    INVALID_FOR_TESTING = 0xFF


class QueryCmdId(GoProEnum):
    """Command ID that is written to GoProUUIDs.CQ_QUERY."""

    GET_SETTING_VAL = 0x12
    GET_STATUS_VAL = 0x13
    GET_SETTING_NAME = 0x22
    GET_CAPABILITIES_VAL = 0x32
    GET_CAPABILITIES_NAME = 0x42
    REG_SETTING_VAL_UPDATE = 0x52
    REG_STATUS_VAL_UPDATE = 0x53
    REG_CAPABILITIES_UPDATE = 0x62
    UNREG_SETTING_VAL_UPDATE = 0x72
    UNREG_STATUS_VAL_UPDATE = 0x73
    UNREG_CAPABILITIES_UPDATE = 0x82
    SETTING_VAL_PUSH = 0x92
    STATUS_VAL_PUSH = 0x93
    SETTING_CAPABILITY_PUSH = 0xA2
    PROTOBUF_QUERY = 0xF5
    INVALID_FOR_TESTING = 0xFF


class StatusId(GoProEnum):
    """Status ID to identify statuses sent to GoProUUIDs.CQ_QUERY or received from GoProUUIDs.CQ_QUERY_RESP."""

    BATT_PRESENT = 1
    BATT_LEVEL = 2
    EXT_BATT_PRESENT = 3
    EXT_BATT_LEVEL = 4
    SYSTEM_HOT = 6
    SYSTEM_BUSY = 8
    QUICK_CAPTURE = 9
    ENCODING = 10
    LCD_LOCK_ACTIVE = 11
    VIDEO_PROGRESS = 13
    INTERNAL_14 = 14
    WIRELESS_ENABLED = 17
    PAIR_STATE = 19
    PAIR_TYPE = 20
    PAIR_TIME = 21
    WAP_SCAN_STATE = 22
    WAP_SCAN_TIME = 23
    WAP_PROV_STAT = 24
    REMOTE_CTRL_VER = 26
    REMOTE_CTRL_CONN = 27
    PAIR_STATE2 = 28
    WLAN_SSID = 29
    AP_SSID = 30
    APP_COUNT = 31
    PREVIEW_ENABLED = 32
    SD_STATUS = 33
    PHOTOS_REM = 34
    VIDEO_REM = 35
    NUM_GROUP_PHOTO = 36
    NUM_GROUP_VIDEO = 37
    NUM_TOTAL_PHOTO = 38
    NUM_TOTAL_VIDEO = 39
    DEPRECATED_40 = 40
    OTA_STAT = 41
    DOWNLOAD_CANCEL_PEND = 42
    MODE_GROUP = 43
    LOCATE_ACTIVE = 45
    INTERNAL_46 = 46
    INTERNAL_47 = 47
    INTERNAL_48 = 48
    MULTI_COUNT_DOWN = 49
    SPACE_REM = 54
    STREAMING_SUPP = 55
    WIFI_BARS = 56
    CURRENT_TIME_MS = 57
    NUM_HILIGHTS = 58
    LAST_HILIGHT = 59
    NEXT_POLL = 60
    ANALYTICS_RDY = 61
    ANALYTICS_SIZE = 62
    IN_CONTEXT_MENU = 63
    TIMELAPSE_REM = 64
    EXPOSURE_TYPE = 65
    EXPOSURE_X = 66
    EXPOSURE_Y = 67
    GPS_STAT = 68
    AP_STATE = 69
    INT_BATT_PER = 70
    ACC_MIC_STAT = 74
    DIGITAL_ZOOM = 75
    WIRELESS_BAND = 76
    DIG_ZOOM_ACTIVE = 77
    MOBILE_VIDEO = 78
    FIRST_TIME = 79
    SEC_SD_STAT = 80
    BAND_5GHZ_AVAIL = 81
    SYSTEM_READY = 82
    BATT_OK_OTA = 83
    CAPTURE_DELAY = 84
    VIDEO_LOW_TEMP = 85
    ORIENTATION = 86
    DEPRECATED_87 = 87
    ZOOM_ENCODING = 88
    FLATMODE_ID = 89
    INTERNAL_90 = 90
    LOGS_READY = 91
    DEPRECATED_92 = 92
    VIDEO_PRESETS = 93
    PHOTO_PRESETS = 94
    TIMELAPSE_PRESETS = 95
    PRESETS_GROUP = 96
    ACTIVE_PRESET = 97
    PRESET_MODIFIED = 98
    LIVE_BURST_REM = 99
    LIVE_BURST_TOTAL = 100
    CAPT_DELAY_ACTIVE = 101
    MEDIA_MOD_MIC_STAT = 102
    TIMEWARP_SPEED_RAMP = 103
    LINUX_CORE_ACTIVE = 104
    CAMERA_LENS_TYPE = 105
    VIDEO_HINDSIGHT = 106
    SCHEDULED_PRESET = 107
    SCHEDULED_CAPTURE = 108
    CREATING_PRESET = 109
    MEDIA_MOD_STAT = 110
    SD_RATING_CHECK_ERROR = 111
    SD_WRITE_SPEED_ERROR = 112
    TURBO_MODE = 113
    CAMERA_CONTROL = 114
    USB_CONNECTED = 115
    CONTROL_OVER_USB = 116
    TOTAL_SD_SPACE_KB = 117


ProducerType = tuple[QueryCmdId, Union[SettingId, StatusId]]
"""Types that can be registered for."""

CmdType = Union[CmdId, QueryCmdId, ActionId]
"""Types that identify a command."""

ResponseType = Union[CmdType, StatusId, SettingId, BleUUID, str, construct.Enum]
"""Types that are used to identify a response."""



# params.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

# pylint: disable=missing-class-docstring, no-member

"""Parameter definitions for GoPro BLE and WiFi commands for Open GoPro version 2.0"""

from __future__ import annotations

from open_gopro.constants import GoProEnum, GoProFlagEnum

# Import required parameters from protobuf
import open_gopro.proto.live_streaming_pb2
import open_gopro.proto.network_management_pb2
import open_gopro.proto.request_get_preset_status_pb2
import open_gopro.proto.set_camera_control_status_pb2


class ScanEntry(GoProFlagEnum):
    AUTHENTICATED = open_gopro.proto.network_management_pb2.SCAN_FLAG_AUTHENTICATED
    CONFIGURED = open_gopro.proto.network_management_pb2.SCAN_FLAG_CONFIGURED
    BEST_SSID = open_gopro.proto.network_management_pb2.SCAN_FLAG_BEST_SSID
    ASSOCIATED = open_gopro.proto.network_management_pb2.SCAN_FLAG_ASSOCIATED
    UNSUPPORTED_TYPE = open_gopro.proto.network_management_pb2.SCAN_FLAG_UNSUPPORTED_TYPE


class ScanState(GoProEnum):
    UNKNOWN = open_gopro.proto.network_management_pb2.SCANNING_UNKNOWN
    NEVER_STARTED = open_gopro.proto.network_management_pb2.SCANNING_NEVER_STARTED
    STARTED = open_gopro.proto.network_management_pb2.SCANNING_STARTED
    ABORTED_BY_SYSTEM = open_gopro.proto.network_management_pb2.SCANNING_ABORTED_BY_SYSTEM
    CANCELLED_BY_USER = open_gopro.proto.network_management_pb2.SCANNING_CANCELLED_BY_USER
    SUCCESS = open_gopro.proto.network_management_pb2.SCANNING_SUCCESS


class LiveStreamStatus(GoProEnum):
    IDLE = open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_IDLE
    CONFIG = open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_CONFIG
    READY = open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_READY
    STREAMING = open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_STREAMING
    STAY_ON_COMPLETE = (
        open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_COMPLETE_STAY_ON
    )
    STAY_ON_FAILED = open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_FAILED_STAY_ON
    RECONNECTING = open_gopro.proto.live_streaming_pb2.EnumLiveStreamStatus.LIVE_STREAM_STATE_RECONNECTING


class LensType(GoProEnum):
    WIDE = open_gopro.proto.live_streaming_pb2.EnumLens.LENS_WIDE
    LINEAR = open_gopro.proto.live_streaming_pb2.EnumLens.LENS_LINEAR
    SUPERVIEW = open_gopro.proto.live_streaming_pb2.EnumLens.LENS_SUPERVIEW


class WindowSize(GoProEnum):
    SIZE_480 = open_gopro.proto.live_streaming_pb2.EnumWindowSize.WINDOW_SIZE_480
    SIZE_720 = open_gopro.proto.live_streaming_pb2.EnumWindowSize.WINDOW_SIZE_720
    SIZE_1080 = open_gopro.proto.live_streaming_pb2.EnumWindowSize.WINDOW_SIZE_1080


class ProvisioningState(GoProEnum):
    UNKNOWN = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_UNKNOWN
    NEVER_STARTED = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_NEVER_STARTED
    STARTED = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_STARTED
    ABORTED_BY_SYSTEM = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_ABORTED_BY_SYSTEM
    CANCELLED_BY_USER = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_CANCELLED_BY_USER
    SUCCESS_NEW_AP = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_SUCCESS_NEW_AP
    SUCCESS_OLD_AP = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_SUCCESS_OLD_AP
    ERROR_FAILED_TO_ASSOCIATE = (
        open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_ERROR_FAILED_TO_ASSOCIATE
    )
    ERROR_PASSWORD_AUTH = (
        open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_ERROR_PASSWORD_AUTH
    )
    ERROR_EULA_BLOCKING = (
        open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_ERROR_EULA_BLOCKING
    )
    ERROR_NO_INTERNET = open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_ERROR_NO_INTERNET
    ERROR_UNSUPPORTED_TYPE = (
        open_gopro.proto.network_management_pb2.EnumProvisioning.PROVISIONING_ERROR_UNSUPPORTED_TYPE
    )


class RegisterPreset(GoProEnum):
    PRESET = (
        open_gopro.proto.request_get_preset_status_pb2.EnumRegisterPresetStatus.REGISTER_PRESET_STATUS_PRESET
    )
    PRESET_GROUP_ARRAY = (
        open_gopro.proto.request_get_preset_status_pb2.EnumRegisterPresetStatus.REGISTER_PRESET_STATUS_PRESET_GROUP_ARRAY
    )


class RegisterLiveStream(GoProEnum):
    MODE = open_gopro.proto.live_streaming_pb2.EnumRegisterLiveStreamStatus.REGISTER_LIVE_STREAM_STATUS_MODE
    ERROR = open_gopro.proto.live_streaming_pb2.EnumRegisterLiveStreamStatus.REGISTER_LIVE_STREAM_STATUS_ERROR
    STATUS = (
        open_gopro.proto.live_streaming_pb2.EnumRegisterLiveStreamStatus.REGISTER_LIVE_STREAM_STATUS_STATUS
    )


class CameraControlStatus(GoProEnum):
    IDLE = open_gopro.proto.set_camera_control_status_pb2.EnumCameraControlStatus.CAMERA_IDLE
    CONTROL = open_gopro.proto.set_camera_control_status_pb2.EnumCameraControlStatus.CAMERA_CONTROL
    EXTERNAL_CONTROL = (
        open_gopro.proto.set_camera_control_status_pb2.EnumCameraControlStatus.CAMERA_EXTERNAL_CONTROL
    )


class PresetGroup(GoProEnum):
    VIDEO = 1000
    PHOTO = 1001
    TIMELAPSE = 1002


class Resolution(GoProEnum):
    NOT_APPLICABLE = 0
    RES_4K = 1
    RES_2_7K = 4
    RES_2_7K_4_3 = 6
    RES_1440 = 7
    RES_1080 = 9
    RES_4K_4_3 = 18
    RES_5K = 24
    RES_5_K_4_3 = 25
    RES_5_3_K_8_7 = 26
    RES_5_3_K_4_3 = 27
    RES_4_K_8_7 = 28
    RES_5_3_K = 100


class FPS(GoProEnum):
    FPS_240 = 0
    FPS_120 = 1
    FPS_100 = 2
    FPS_60 = 5
    FPS_50 = 6
    FPS_30 = 8
    FPS_25 = 9
    FPS_24 = 10
    FPS_200 = 13


class AutoOff(GoProEnum):
    NEVER = 0
    MIN_5 = 4
    MIN_15 = 6
    MIN_30 = 7


class LensMode(GoProEnum):
    SINGLE = 0
    DUAL = 5


class VideoFOV(GoProEnum):
    WIDE = 0
    NARROW = 2
    SUPERVIEW = 3
    LINEAR = 4
    MAX_SUPERVIEW = 7
    LINEAR_HORIZON_LEVELING = 8
    HYPERVIEW = 9
    LINEAR_HORIZON_LOCK = 10


class PhotoFOV(GoProEnum):
    NOT_APPLICABLE = 0
    HYPERVIEW = 9
    NARROW = 19
    WIDE = 101
    LINEAR = 102
    MAX_SUPERVIEW = 100
    LINEAR_HORIZON = 121


class MultishotFOV(GoProEnum):
    NOT_APPLICABLE = 0
    NARROW = 19
    MAX_SUPERVIEW = 100
    WIDE = 101
    LINEAR = 102


class LED(GoProEnum):
    LED_2 = 2
    ALL_ON = 3
    ALL_OFF = 4
    FRONT_OFF_ONLY = 5
    BLE_KEEP_ALIVE = 66


class PairState(GoProEnum):
    NOT_STARTED = 0
    IN_PROGRESS = 1
    FAILED = 2
    STOPPED = 3
    COMPLETED = 4


class PairType(GoProEnum):
    NOT_PAIRING = 0
    PAIRING_APP = 1
    PAIRING_REMOTE_CONTROL = 2
    PAIRING_BLUETOOTH = 3


class WAPState(GoProEnum):
    NEVER_STARTED = 0
    STARTED = 1
    ABORTED = 2
    CANCELED = 3
    COMPLETED = 4


class SDStatus(GoProEnum):
    OK = 0
    FULL = 1
    REMOVED = 2
    FORMAT_ERROR = 3
    BUSY = 4
    SWAPPED = 8
    UNKNOWN = 0xFF


class OTAStatus(GoProEnum):
    IDLE = 0
    DOWNLOADING = 1
    VERIFYING = 2
    DOWNLOAD_FAILED = 3
    VERIFY_FAILED = 4
    READY = 5
    GOPRO_APP_DOWNLOADING = 6
    GOPRO_APP_VERIFYING = 7
    GOPRO_APP_DOWNLOAD_FAILED = 8
    GOPRO_APP_VERIFY_FAILED = 9
    GOPRO_APP_READY = 10


class AnalyticsState(GoProEnum):
    NOT_READY = 0
    READY = 1
    ON_CONNECT = 2


class ExposureMode(GoProEnum):
    DISABLED = 0
    AUTO = 1
    ISO_LOCK = 2
    HEMISPHERE = 3


class AccMicStatus(GoProEnum):
    NOT_CONNECTED = 0
    CONNECTED = 1
    CONNECTED_AND_PLUGGED_IN = 2


class WifiBand(GoProEnum):
    BAND_2_4_GHZ = 0
    BAND_5_GHZ = 1
    BAND_MAX = 2


class Orientation(GoProEnum):
    UPRIGHT = 0
    UPSIDE_DOWN = 1
    ON_RIGHT_SIDE = 2
    ON_LEFT_SIDE = 3


class MediaModMicStatus(GoProEnum):
    REMOVED = 0
    ONLY = 1
    WITH_EXTERNAL = 2


class TimeWarpSpeed(GoProEnum):
    SPEED_15X = 0
    SPEED_30X = 1
    SPEED_60X = 2
    SPEED_150X = 3
    SPEED_300X = 4
    SPEED_900X = 5
    SPEED_1800X = 6
    SPEED_2X = 7
    SPEED_5X = 8
    SPEED_10X = 9
    SPEED_AUTO = 10
    SPEED_1X = 11
    SPEED_HALF = 12


class MaxLensMode(GoProEnum):
    DEFAULT = 0
    MAX_LENS = 1


class MediaModStatus(GoProEnum):
    SELFIE_0_HDMI_0_MEDIAMODCONNECTED_FALSE = 0
    SELFIE_0_HDMI_0_MEDIAMODCONNECTED_TRUE = 1
    SELFIE_0_HDMI_1_MEDIAMODCONNECTED_FALSE = 2
    SELFIE_0_HDMI_1_MEDIAMODCONNECTED_TRUE = 3
    SELFIE_1_HDMI_0_MEDIAMODCONNECTED_FALSE = 4
    SELFIE_1_HDMI_0_MEDIAMODCONNECTED_TRUE = 5
    SELFIE_1_HDMI_1_MEDIAMODCONNECTED_FALSE = 6
    SELFIE_1_HDMI_1_MEDIAMODCONNECTED_TRUE = 7


class Flatmode(GoProEnum):
    VIDEO = 12
    LOOPING = 15
    SINGLE_PHOTO = 16
    NIGHT_PHOTO = 18
    BURST_PHOTO = 19
    TIME_LAPSE_VIDEO = 13
    TIME_LAPSE_PHOTO = 20
    NIGHT_LAPSE_PHOTO = 21
    TIME_WARP_VIDEO = 24
    LIVE_BURST = 25
    NIGHT_LAPSE_VIDEO = 26
    SLO_MO = 27


class Toggle(GoProEnum):
    ENABLE = 1
    DISABLE = 0


class HypersmoothMode(GoProEnum):
    OFF = 0
    UNKNOWN = 1
    HIGH = 2
    BOOST = 3
    AUTO_BOOST = 4
    STANDARD = 100


class CameraControl(GoProEnum):
    IDLE = 0
    CAMERA = 1
    EXTERNAL = 2


class PerformanceMode(GoProEnum):
    MAX_PERFORMANCE = 0
    EXTENDED_BATTERY = 1
    STATIONARY = 2


class MediaFormat(GoProEnum):
    TODO_12 = 12
    TIME_LAPSE_VIDEO = 13
    TODO_16 = 16
    TIME_LAPSE_PHOTO = 20
    NIGHT_LAPSE_PHOTO = 21
    NIGHT_LAPSE_VIDEO = 26


class AntiFlicker(GoProEnum):
    HZ_60 = 2
    HZ_50 = 3


class CameraUxMode(GoProEnum):
    EASY = 0
    PRO = 1


class HorizonLeveling(GoProEnum):
    OFF = 0
    LOCKED = 2


class Speed(GoProEnum):
    ULTRA_SLO_MO_8X = 0
    SUPER_SLO_MO_4X = 1
    SLO_MO_2X = 2
    LOW_LIGHT_1X = 3
    SUPER_SLO_MO_4X_EXT_BATT = 4
    SLO_MO_2X_EXT_BATT = 5
    LOW_LIGHT_1X_EXT_BATT = 6
    ULTRA_SLO_MO_8X_50_HZ = 7
    SUPER_SLO_MO_4X_50_HZ = 8
    SLO_MO_2X_50_HZ = 9
    LOW_LIGHT_1X_50_HZ = 10
    SUPER_SLO_MO_4X_EXT_BATT_50_HZ = 11
    SLO_MO_2X_EXT_BATT_50_HZ = 12
    LOW_LIGHT_1X_EXT_BATT_50_HZ = 13


class PhotoEasyMode(GoProEnum):
    OFF = 0
    ON = 1
    SUPER_PHOTO = 100
    NIGHT_PHOTO = 101


class StarTrailLength(GoProEnum):
    NOT_APPLICABLE = 0
    SHORT = 1
    LONG = 2
    MAX = 3


class SystemVideoMode(GoProEnum):
    HIGHEST_QUALITY = 0
    EXTENDED_BATTERY = 1



# preset_control.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Aug 30 17:53:34 UTC 2022

"""Entrypoint for taking a picture demo."""

from __future__ import annotations
import enum
import logging
import argparse
from typing import Optional, Union, Callable

from rich.console import Console
from rich.prompt import IntPrompt, Prompt
from rich.table import Table
from rich import box

from open_gopro import GoPro, GoProResp
from open_gopro import Params as GoProParams
from open_gopro.constants import GoProEnum, SettingId
from open_gopro.util import set_stream_logging_level, setup_logging, add_cli_args_and_parse

console = Console()  # rich consoler printer
gopro: Optional[GoPro] = None

######### BEGIN PRESET PARSER #########


class Setting:
    """Class representation of preset setting

    Data populated by parsing preset status response
    """

    def __init__(self, settingDict: dict):
        """Constructor

        Raises:
            RuntimeError: Error parsing Setting from response dict

        Args:
            settingDict (dict): dict representation of setting
        """
        self.id: Union[SettingId, int]
        self.value: Union[GoProEnum, int]
        self.is_caption: bool

        valueType: Optional[Callable] = None
        if (parsedId := settingDict.get("id")) is not None:
            if parsedId in [i.value for i in SettingId]:
                self.id = SettingId(parsedId)
                valueType = GoProResp._get_query_container(self.id)
            else:
                self.id = parsedId
        else:
            raise RuntimeError("Error parsing preset id for Setting")
        if (parsedValue := settingDict.get("value")) is not None:
            self.value = valueType(parsedValue) if valueType else parsedValue
        else:
            raise RuntimeError("Error parsing preset value for Setting")
        if (parsedis_caption := settingDict.get("is_caption")) is not None:
            self.is_caption = parsedis_caption
        else:
            raise RuntimeError("Error parsing preset is_caption for Setting")

    def to_table(self) -> Table:
        """Formats data into rich table

        Returns:
            Table: rich table representation of data
        """
        settingTable = Table("Setting", box=box.SQUARE)

        settingTable.add_row("id:", str(self.id))
        settingTable.add_row("value:", str(self.value))
        settingTable.add_row("is_caption:", str(self.is_caption))

        return settingTable


class Preset:
    """Class representation of preset

    Data populated by parsing preset status response
    """

    def __init__(self, presetDict: dict):
        """Constructor

        Raises:
            RuntimeError: Error parsing Setting from response dict

        Args:
            presetDict (dict): dict representation of preset
        """
        self.id: int
        self.user_defined: bool = False
        self.is_modified: bool = False
        self.mode: Optional[enum.Enum] = None
        self.title_id: Optional[enum.Enum] = None
        self.icon: Optional[enum.Enum] = None
        self.is_fixed: Optional[bool] = None
        self.setting_array: Optional[list[Setting]] = None

        if (parsedId := presetDict.get("id")) is not None:
            self.id = parsedId
        else:
            raise RuntimeError("Error parsing preset id for Preset")

        if (parsed_user_defined := presetDict.get("user_defined")) is not None:
            self.user_defined = parsed_user_defined
        else:
            raise RuntimeError("Error parsing user_defined for Preset")

        if (parsed_is_modified := presetDict.get("is_modified")) is not None:
            self.is_modified = parsed_is_modified
        else:
            raise RuntimeError("Error parsing is_modified for Preset")

        if (parsedMode := presetDict.get("mode")) is not None:
            self.mode = parsedMode
        if (parsed_title_id := presetDict.get("title_id")) is not None:
            self.title_id = parsed_title_id
        if (parsed_icon := presetDict.get("icon")) is not None:
            self.icon = parsed_icon
        if (parsed_is_fixed := presetDict.get("is_fixed")) is not None:
            self.is_fixed = parsed_is_fixed
        if (settingDictArray := presetDict.get("setting_array")) is not None:
            if isinstance(settingDictArray, list):
                self.setting_array = []
                for settingDict in settingDictArray:
                    self.setting_array.append(Setting(settingDict))

    def to_table(self, activePreset: Optional[int] = None) -> Table:
        """Formats data into rich table

        Args:
            activePreset (Union[GoProParams.Preset, int, None]): The currently active preset if known, used to highlight active preset in table

        Returns:
            Table: rich table representation of data
        """
        presetTable = Table("Preset", box=box.SQUARE)
        if (activePreset is not None) and activePreset == self.id:
            presetTable.add_column("ACTIVE")
            presetTable.box = box.HEAVY
        else:
            presetTable.add_column("")

        presetTable.add_row("id:", str(self.id))
        presetTable.add_row("user_defined:", str(self.user_defined))

        presetTable.add_row("is_modified:", str(self.is_modified))

        if self.mode is not None:
            presetTable.add_row("mode:", str(self.mode))
        if self.title_id is not None:
            presetTable.add_row("title_id:", str(self.title_id))
        if self.icon is not None:
            presetTable.add_row("icon:", str(self.icon))
        if self.is_fixed is not None:
            presetTable.add_row("is_fixed:", str(self.is_fixed))
        if self.setting_array is not None:
            settingTables: list[Table] = []
            for setting in self.setting_array:
                settingTables.append(setting.to_table())
            # Use horizontal grid layout for compactness
            setting_arrayGrid = Table.grid(expand=False)
            setting_arrayGrid.add_row(*settingTables)
            presetTable.add_row("Settings:", setting_arrayGrid)

        return presetTable


class PresetGroup:
    """Class representation of preset group

    Data populated by parsing preset status response
    """

    def __init__(self, presetGroupDict: dict):
        """Constructor

        Raises:
            RuntimeError: Error parsing Setting from response dict

        Args:
            presetGroupDict (dict): dict representation of preset group
        """
        self.id: Union[GoProParams.PresetGroup, int]
        self.presetArray: list[Preset] = []
        self.can_add_preset: Optional[bool] = None

        # Parse id
        if (parsedId := presetGroupDict.get("id")) is not None:
            self.id = parsedId
        else:
            raise RuntimeError("Error parsing id for PresetGroup")

        # Parse can_add_preset
        if (parsed_can_add_preset := presetGroupDict.get("can_add_preset")) is not None:
            self.can_add_preset = parsed_can_add_preset

        # Parse preset_array
        presetDictArray: Optional[list[dict]] = presetGroupDict.get("preset_array")
        if presetDictArray is not None:
            for preset in presetDictArray:
                self.presetArray.append(Preset(preset))
        else:
            raise RuntimeError("Error parsing presetArray for PresetGroup")

    def to_table(self, activePreset: Optional[int] = None) -> Table:
        """Formats data into rich table

        Args:
            activePreset (Optional[int]): The currently active preset if known, used to highlight active preset in table

        Returns:
            Table: rich table representation of data
        """
        presetGroupTable = Table("Preset Group", box=box.SQUARE)

        presetGroupTable.add_row("id:", str(self.id))

        if self.can_add_preset is not None:
            presetGroupTable.add_row("can_add_preset:", str(self.can_add_preset))

        # Horizontal layout is more compact but can cause present and setting attributes to be cut off
        for preset in self.presetArray:
            presetGroupTable.add_row("", preset.to_table(activePreset))

        return presetGroupTable

    def get_preset_by_id(self, presetId: int) -> Optional[Preset]:
        """helper function for fetching the preset with specified preset id

        Args:
            presetId (int): The preset id being searched for

        Returns:
            Optional[Preset]: The matching preset found
        """
        for preset in self.presetArray:
            if presetId == preset.id:
                return preset
        return None


class PresetCollection:
    """Class representation of preset collection

    Data populated by parsing preset status response
    """

    def __init__(self, presetCollectionDict: dict):
        """Constructor

        Raises:
            RuntimeError: Error parsing Setting from response dict

        Args:
            presetCollectionDict (dict): dict representation of preset collection
        """
        self.presetGroupArray: list[PresetGroup] = []

        # Parse presetGroupArray
        presetGroupDictArray = presetCollectionDict.get("preset_group_array")
        if isinstance(presetGroupDictArray, list):
            for presetGroupDict in presetGroupDictArray:
                self.presetGroupArray.append(PresetGroup(presetGroupDict))
        else:
            raise RuntimeError("Error parsing presetGroupArray for PresetCollection")

    #
    def to_table(self, activePreset: Optional[int] = None) -> Table:
        """Formats data into rich table

        Args:
            activePreset (Union[GoProParams.Preset, int, None]): The currently active preset if known, used to
                highlight active preset in table

        Returns:
            Table: rich table representation of data
        """
        presetCollectionTable = Table("Preset Collection", box=box.SQUARE)

        for presetGroup in self.presetGroupArray:
            presetCollectionTable.add_row(presetGroup.to_table(activePreset))

        return presetCollectionTable

    def get_preset_tuple(self, presetId: int) -> tuple[Optional[PresetGroup], Optional[Preset]]:
        """helper function for fetching the preset with specified preset id and its parent preset group

        Args:
            presetId (int): The preset id being searched for

        Returns:
            tuple[Optional[PresetGroup], Optional[Preset]]: The matching preset and preset group if found
        """
        for presetGroup in self.presetGroupArray:
            if preset := presetGroup.get_preset_by_id(presetId):
                return presetGroup, preset
        return None, None


######### END PRESET PARSER #########


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Connect to a GoPro camera's Wifi Access Point.")
    return add_cli_args_and_parse(parser)


class MenuMode(enum.Enum):
    """Enum to identify the current menu state"""

    MAIN = 0
    TOP_LEVEL_SETTINGS = 1
    COLLECTION = 2
    DEMO = 3


class CameraModel(enum.Enum):
    """Enum to identify the camera model"""

    HERO10 = 0
    HERO11 = 1
    HERO11_MINI = 2


def applySettingChange(setting: SettingId, value: GoProEnum) -> None:
    """Applies specified value to the specified setting

    Args:
        setting (SettingId): The setting to modify
        value (GoProEnum): The value to apply to the setting
    """
    if gopro is None:
        return

    if setting not in gopro.ble_setting:
        return

    gopro.ble_setting[setting].set(value)


def printCurrentPresetStatus() -> None:
    """Prints the current preset status"""
    if gopro is None:
        return

    presetCollection = PresetCollection(gopro.ble_command.get_preset_status().data)
    activePresetId = gopro.ble_status.active_preset.get_value().flatten
    console.print(presetCollection.to_table(activePresetId))


def presetDemo(cameraModel: CameraModel) -> None:
    """Demonstrates setting combinations which cause distinct preset collections

    Args:
        cameraModel (CameraModel): Model for the paired camera
    """
    if gopro is None:
        return

    console.print(
        "[green]This demo walks through all setting combinations that cause distinct preset combinations."
    )

    if cameraModel == CameraModel.HERO11_MINI:
        # Easy
        console.print("[blue bold]Preset Collection 1")
        console.print("[gray]UX Mode:[/gray] [purple]Easy")
        applySettingChange(SettingId.CAMERA_UX_MODE, GoProParams.CameraUxMode.EASY)
        printCurrentPresetStatus()
        Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

        # Pro
        console.print("[blue bold]Preset Collection 2")
        console.print("[gray]UX Mode:[/gray] [purple]Pro")
        applySettingChange(SettingId.CAMERA_UX_MODE, GoProParams.CameraUxMode.PRO)
        printCurrentPresetStatus()
        Prompt.ask(prompt="[yellow]Press enter to end demo")

    else:
        # Max Lens
        console.print("[blue bold]Preset Collection 1")
        console.print("[gray]Max Lens:[/gray] [purple]On")
        applySettingChange(SettingId.MAX_LENS_MOD, GoProParams.MaxLensMode.MAX_LENS)
        printCurrentPresetStatus()
        Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

        if cameraModel == CameraModel.HERO11:
            # Easy Highest Quality
            console.print("[blue bold]Preset Collection 2")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]UX Mode:[/gray] [purple]Easy")
            console.print("[gray]Easy Night Photo:[/gray] [purple]Off")
            console.print("[gray]System Video Mode:[/gray] [purple]Highest Quality")
            applySettingChange(SettingId.MAX_LENS_MOD, GoProParams.MaxLensMode.DEFAULT)
            applySettingChange(SettingId.CAMERA_UX_MODE, GoProParams.CameraUxMode.EASY)
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.PHOTO)
            applySettingChange(SettingId.PHOTO_EASY_MODE, GoProParams.PhotoEasyMode.OFF)
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.VIDEO)
            applySettingChange(SettingId.SYSTEM_VIDEO_MODE, GoProParams.SystemVideoMode.HIGHEST_QUALITY)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Easy Highest Quality Night Photo
            console.print("[blue bold]Preset Collection 3")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]UX Mode:[/gray] [purple]Easy")
            console.print("[gray]Easy Night Photo:[/gray] [purple]On")
            console.print("[gray]System Video Mode:[/gray] [purple]Highest Quality")
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.PHOTO)
            applySettingChange(SettingId.PHOTO_EASY_MODE, GoProParams.PhotoEasyMode.NIGHT_PHOTO)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Easy Extended Battery
            console.print("[blue bold]Preset Collection 4")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]UX Mode:[/gray] [purple]Easy")
            console.print("[gray]Easy Night Photo:[/gray] [purple]Off")
            console.print("[gray]System Video Mode:[/gray] [purple]Extended Battery")
            applySettingChange(SettingId.PHOTO_EASY_MODE, GoProParams.PhotoEasyMode.OFF)
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.VIDEO)
            applySettingChange(SettingId.SYSTEM_VIDEO_MODE, GoProParams.SystemVideoMode.EXTENDED_BATTERY)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Easy Extended Battery Night Photo
            console.print("[blue bold]Preset Collection 5")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]UX Mode:[/gray] [purple]Easy")
            console.print("[gray]Easy Night Photo:[/gray] [purple]On")

            console.print("[gray]System Video Mode:[/gray] [purple]Extended Battery")
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.PHOTO)
            applySettingChange(SettingId.PHOTO_EASY_MODE, GoProParams.PhotoEasyMode.NIGHT_PHOTO)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Pro Highest Quality
            console.print("[blue bold]Preset Collection 6")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]UX Mode:[/gray] [purple]Pro")
            console.print("[gray]System Video Mode:[/gray] [purple]Highest Quality")
            applySettingChange(SettingId.CAMERA_UX_MODE, GoProParams.CameraUxMode.PRO)
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.VIDEO)
            applySettingChange(SettingId.SYSTEM_VIDEO_MODE, GoProParams.SystemVideoMode.EXTENDED_BATTERY)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Pro Extended Battery
            console.print("[blue bold]Preset Collection 7")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]UX Mode:[/gray] [purple]Pro")
            console.print("[gray]System Video Mode:[/gray] [purple]Extended Battery")
            applySettingChange(SettingId.CAMERA_UX_MODE, GoProParams.CameraUxMode.PRO)
            gopro.ble_command.load_preset_group(GoProParams.PresetGroup.VIDEO)
            applySettingChange(SettingId.SYSTEM_VIDEO_MODE, GoProParams.SystemVideoMode.EXTENDED_BATTERY)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to end demo")
        else:
            # Max Performance
            console.print("[blue bold]Preset Collection 2")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]Video Performance Mode:[/gray] [purple]Max Performance")
            applySettingChange(SettingId.MAX_LENS_MOD, GoProParams.MaxLensMode.DEFAULT)
            applySettingChange(SettingId.VIDEO_PERFORMANCE_MODE, GoProParams.PerformanceMode.MAX_PERFORMANCE)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Extended Battery
            console.print("[blue bold]Preset Collection 3")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]Video Performance Mode:[/gray] [purple]Extended Battery")
            applySettingChange(SettingId.VIDEO_PERFORMANCE_MODE, GoProParams.PerformanceMode.EXTENDED_BATTERY)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to continue to the next preset collection")

            # Stationary
            console.print("[blue bold]Preset Collection 4")
            console.print("[gray]Max Lens:[/gray] [purple]Off")
            console.print("[gray]Video Performance Mode:[/gray] [purple]Stationary")
            applySettingChange(SettingId.VIDEO_PERFORMANCE_MODE, GoProParams.PerformanceMode.STATIONARY)
            printCurrentPresetStatus()
            Prompt.ask(prompt="[yellow]Press enter to end demo")


def main(args: argparse.Namespace) -> None:
    global gopro
    logger = setup_logging(__name__, args.log)

    # Constants
    topLevelSettings: list[SettingId]
    hero10_topLevelSettings: list[SettingId] = [SettingId.MAX_LENS_MOD, SettingId.VIDEO_PERFORMANCE_MODE]
    hero11_topLevelSettings: list[SettingId] = [
        SettingId.MAX_LENS_MOD,
        SettingId.SYSTEM_VIDEO_MODE,
        SettingId.CAMERA_UX_MODE,
        SettingId.PHOTO_EASY_MODE,
    ]
    hero11_mini_topLevelSettings: list[SettingId] = [SettingId.CAMERA_UX_MODE]

    # State
    collectionCacheInvalidated = True
    activeCacheInvalidated = True
    loopRunning = True
    menuMode: MenuMode = MenuMode.MAIN
    focusedPresetGroup: Optional[PresetGroup] = None
    focusedPreset: Optional[Preset] = None
    focusedSetting: Optional[Setting] = None
    focusedTopLevelSetting: Optional[SettingId] = None
    activePresetGroup: Optional[PresetGroup] = None
    activePreset: Optional[Preset] = None
    presetCollectionCache: Optional[PresetCollection] = None
    cameraModel: CameraModel = CameraModel.HERO10

    with GoPro(
        args.identifier,
        wifi_interface=args.wifi_interface,
        sudo_password=args.password,
        enable_wifi=False,
    ) as gopro:
        # Now we only want errors
        set_stream_logging_level(logging.ERROR)

        if (fwVersion := gopro.ble_command.get_hardware_info().data.get("firmware_version")) is not None:
            if fwVersion.split(".")[0] == "H21":
                cameraModel = CameraModel.HERO10
            elif fwVersion.split(".")[0] == "H22":
                if fwVersion.split(".")[1] == "01":
                    cameraModel = CameraModel.HERO11
                else:
                    cameraModel = CameraModel.HERO11_MINI

        if cameraModel == CameraModel.HERO10:
            console.print("Detected Hero 10")
            topLevelSettings = hero10_topLevelSettings
        elif cameraModel == CameraModel.HERO11:
            console.print("Detected Hero 11 camera")
            topLevelSettings = hero11_topLevelSettings
        elif cameraModel == CameraModel.HERO11_MINI:
            console.print("Detected Hero 11 Mini camera")
            topLevelSettings = hero11_mini_topLevelSettings
        else:
            console.print("Unsupported camera detected")
            return

        while loopRunning:
            # REFRESH INVALID CACHES
            if collectionCacheInvalidated:
                console.print("[yellow]Refreshing preset status cache")
                presetCollectionCache = PresetCollection(
                    gopro.ble_command.get_preset_status(
                        register_preset_status=[GoProParams.RegisterPreset.PRESET]
                    ).data
                )
                collectionCacheInvalidated = False

            if presetCollectionCache is None:
                logger.error("Unable to fetch or parse current preset status, ending demo")
                loopRunning = False
                continue

            if activeCacheInvalidated:
                console.print("[yellow]Refreshing active preset cache")
                activePresetId = gopro.ble_status.active_preset.get_value().flatten
                activePresetGroup, activePreset = presetCollectionCache.get_preset_tuple(activePresetId)
                activeCacheInvalidated = False

            # MAIN MENU
            if menuMode == MenuMode.MAIN:
                console.print("[bold blue]Main Menu")
                menu = Table.grid(expand=False)
                menu.add_column(justify="right", width=4)
                menu.add_column()
                menu.add_column()
                menu.add_row("0", " ) ", "Print current preset status")
                menu.add_row("1", " ) ", "Modify top level settings")
                menu.add_row("2", " ) ", "Modify preset collection settings")
                menu.add_row("3", " ) ", "Run Preset Availability Demo")
                menu.add_row("4", " ) ", "Exit")
                console.print(menu)
                option = IntPrompt.ask(
                    prompt="Select an option from the menu above",
                    choices=[str(i) for i in range(0, 5)],
                    show_choices=False,
                )
                if option == 0:
                    if activePreset is not None:
                        console.print(presetCollectionCache.to_table(activePreset.id))
                    else:
                        console.print(presetCollectionCache.to_table())
                elif option == 1:
                    menuMode = MenuMode.TOP_LEVEL_SETTINGS
                    focusedTopLevelSetting = None
                elif option == 2:
                    menuMode = MenuMode.COLLECTION
                    focusedPresetGroup = None
                    focusedPreset = None
                    focusedSetting = None
                elif option == 3:
                    menuMode = MenuMode.DEMO
                else:
                    loopRunning = False

            # TOP LEVEL SETTING MENU
            elif menuMode == MenuMode.TOP_LEVEL_SETTINGS:
                if focusedTopLevelSetting is None:
                    menu = Table.grid(expand=False)
                    menu.add_column(justify="right", width=4)
                    menu.add_column()
                    menu.add_column()
                    menu.add_row("0", " ) ", "Back to Main Menu")
                    option_offset = 1
                    for i, settingId in enumerate(topLevelSettings):
                        if settingId not in gopro.ble_setting:
                            menu.add_row(
                                str(i + option_offset),
                                " ) ",
                                "(Inaccessible) Setting " + str(settingId),
                            )
                        else:
                            menu.add_row(str(i + option_offset), " ) ", "Modify setting " + str(settingId))

                    console.print("[bold blue]Top Level Setting Menu")
                    console.print(menu)
                    option = IntPrompt.ask(
                        prompt="Select an option from the menu above",
                        choices=[str(i) for i in range(len(topLevelSettings) + option_offset)],
                        show_choices=False,
                    )
                    if option == 0:
                        menuMode = MenuMode.MAIN
                    if topLevelSettings[option - option_offset] not in gopro.ble_setting:
                        console.print("[red]Invalid selection")
                    else:
                        focusedTopLevelSetting = topLevelSettings[option - option_offset]
                else:
                    menu = Table.grid(expand=False)
                    menu.add_column(justify="right", width=4)
                    menu.add_column()
                    menu.add_column()
                    menu.add_row("0", " ) ", "Back to Top Level Settings Menu")
                    option_offset = 1

                    try:
                        settingHandler = gopro.ble_setting[focusedTopLevelSetting]

                        logger.debug(f"Fetching value for setting {str(focusedTopLevelSetting)}")
                        response = settingHandler.get_value()
                        logger.debug(response)

                        currentValue = response.data.get(focusedTopLevelSetting)

                        logger.debug(f"Fetching capabilities for setting {str(focusedTopLevelSetting)}")
                        response = settingHandler.get_capabilities_values()
                        logger.debug(response)

                        settingValues = response.data.get(focusedTopLevelSetting)
                        if settingValues is None:
                            console.print(
                                f"No reported capabilities, cannot edit setting {str(focusedTopLevelSetting)}"
                            )
                            focusedTopLevelSetting = None
                        else:
                            for i, value in enumerate(settingValues):
                                if value == currentValue:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Capability " + str(value) + " [green](CURRENT)",
                                    )
                                else:
                                    menu.add_row(str(i + option_offset), " ) ", "Capability " + str(value))

                            console.print(f"[bold blue]Setting Menu: ({str(focusedTopLevelSetting)})")
                            console.print(menu)
                            option = IntPrompt.ask(
                                prompt="Select an option from the menu above",
                                choices=[str(i) for i in range(len(settingValues) + option_offset)],
                                show_choices=False,
                            )
                            if option == 0:
                                focusedTopLevelSetting = None
                            else:
                                console.print(
                                    f"Changing setting {str(focusedTopLevelSetting)} value to {str(settingValues[option - option_offset])}"
                                )
                                logger.debug(
                                    f"Changing setting {str(focusedTopLevelSetting)} to {str(settingValues[option - option_offset])}"
                                )
                                response = settingHandler.set(settingValues[option - option_offset])
                                logger.debug(response)
                                collectionCacheInvalidated = True
                                activeCacheInvalidated = True
                    except KeyError:
                        console.print(
                            f"Unable to edit setting, no setting handler found for {str(focusedTopLevelSetting)}"
                        )
                        focusedTopLevelSetting = None
            # COLLECTION MENU
            elif menuMode == MenuMode.COLLECTION:
                if focusedPresetGroup is None:
                    menu = Table.grid(expand=False)
                    menu.add_column(justify="right", width=4)
                    menu.add_column()
                    menu.add_column()
                    menu.add_column()
                    menu.add_row("0", " ) ", "Back to Main Menu")
                    option_offset = 1
                    if isinstance(presetCollectionCache.presetGroupArray, list):
                        for i, group in enumerate(presetCollectionCache.presetGroupArray):
                            if group == activePresetGroup:
                                menu.add_row(
                                    str(i + option_offset),
                                    " ) ",
                                    "Preset Group " + str(group.id) + " [green](ACTIVE)",
                                )
                            else:
                                menu.add_row(
                                    str(i + option_offset),
                                    " ) ",
                                    "Preset Group " + str(group.id),
                                )

                        console.print("[bold blue]Preset Collection Menu")
                        console.print(menu)
                        option = IntPrompt.ask(
                            prompt="Select an option from the menu above",
                            choices=[
                                str(i)
                                for i in range(len(presetCollectionCache.presetGroupArray) + option_offset)
                            ],
                            show_choices=False,
                        )
                        if option == 0:
                            menuMode = MenuMode.MAIN
                        else:
                            focusedPresetGroup = presetCollectionCache.presetGroupArray[option - option_offset]

                elif focusedPreset is None:
                    menu = Table.grid(expand=False)
                    menu.add_column(justify="right", width=4)
                    menu.add_column()
                    menu.add_column()
                    menu.add_column()
                    menu.add_row("0", " ) ", "Back to Preset Collection Menu")
                    option_offset = 1
                    if isinstance(focusedPresetGroup.presetArray, list):
                        for i, preset in enumerate(focusedPresetGroup.presetArray):
                            if preset.title_id is None:
                                if preset == activePreset:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Preset " + str(preset.id) + " [green](ACTIVE)",
                                    )
                                else:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Preset " + str(preset.id),
                                    )
                            else:
                                if preset == activePreset:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Preset " + str(preset.title_id) + " [green](ACTIVE)",
                                    )
                                else:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Preset " + str(preset.title_id),
                                    )

                        console.print(f"[bold blue]Preset Group Menu: ({str(focusedPresetGroup.id)})")

                        console.print(menu)
                        option = IntPrompt.ask(
                            prompt="Select an option from the menu above",
                            choices=[
                                str(i) for i in range(len(focusedPresetGroup.presetArray) + option_offset)
                            ],
                            show_choices=False,
                        )
                        if option == 0:
                            focusedPresetGroup = None
                        elif (
                            focusedPresetGroup.presetArray[option - option_offset].setting_array is None
                        ) or (focusedPresetGroup.presetArray[option - option_offset].title_id is None):
                            console.print("[red]Invalid selection")
                        else:
                            focusedPreset = focusedPresetGroup.presetArray[option - option_offset]

                elif focusedSetting is None:
                    menu = Table.grid(expand=False)
                    menu.add_column(justify="right", width=4)
                    menu.add_column()
                    menu.add_column()
                    menu.add_row("0", " ) ", "Back to Preset Group Menu")
                    option_offset = 1

                    if focusedPreset != activePreset:
                        menu.add_row("1", " ) ", "Load this preset")
                        option_offset += 1
                    else:
                        if isinstance(focusedPreset.setting_array, list):
                            for i, setting in enumerate(focusedPreset.setting_array):
                                if setting.id in gopro.ble_setting:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Setting " + str(setting.id),
                                    )
                                else:
                                    menu.add_row(
                                        str(i + option_offset),
                                        " ) ",
                                        "Setting " + str(setting.id) + " [red](NO API SUPPORT)",
                                    )

                    if focusedPreset == activePreset:
                        console.print(
                            f"[bold blue]Preset Menu: ({str(focusedPreset.title_id)})[/bold blue] [green](ACTIVE)"
                        )
                    else:
                        console.print(f"[bold blue]Preset Menu: ({str(focusedPreset.title_id)})")
                    console.print(menu)
                    optionCount = option_offset
                    if focusedPreset.setting_array is not None:
                        optionCount += len(focusedPreset.setting_array)
                    option = IntPrompt.ask(
                        prompt="Select an option from the menu above",
                        choices=[str(i) for i in range(optionCount)],
                        show_choices=False,
                    )
                    if option == 0:
                        focusedPreset = None
                    elif option == 1 and (focusedPreset != activePreset):
                        if focusedPreset.title_id is None:
                            console.print(f"[purple]Loading preset {str(focusedPreset.id)}")
                        else:
                            console.print(f"[purple]Loading preset {str(focusedPreset.title_id)}")
                        gopro.ble_command.load_preset(focusedPreset.id)
                        activeCacheInvalidated = True
                    elif isinstance(focusedPreset.setting_array, list):
                        if focusedPreset.setting_array[option - option_offset].id not in gopro.ble_setting:
                            console.print("[red]Invalid selection")
                        else:
                            focusedSetting = focusedPreset.setting_array[option - option_offset]

                else:
                    menu = Table.grid(expand=False)

                    menu.add_column(justify="right", width=4)
                    menu.add_column()
                    menu.add_column()
                    menu.add_row("0", " ) ", "Back to Preset Menu")
                    option_offset = 1

                    try:
                        if not isinstance(focusedSetting.id, SettingId):
                            raise KeyError

                        settingHandler = gopro.ble_setting[focusedSetting.id]
                    except KeyError:
                        console.print(
                            f"Unable to edit setting, no setting handler found for {str(focusedSetting.id)}"
                        )
                        focusedSetting = None
                        return

                    logger.debug("Fetching capabilities for setting " + str(focusedSetting.id))
                    response = settingHandler.get_capabilities_values()
                    logger.debug(response)

                    settingValues = response.data.get(focusedSetting.id)

                    if not isinstance(settingValues, list):
                        console.print(
                            f"Failed to retrieve capabilities, cannot edit setting {str(focusedSetting.id)}"
                        )
                        focusedSetting = None
                    else:
                        for i, value in enumerate(settingValues):
                            if value == focusedSetting.value:
                                menu.add_row(
                                    str(i + option_offset),
                                    " ) ",
                                    "Capability " + str(value) + " [green](CURRENT)",
                                )
                            else:
                                menu.add_row(str(i + option_offset), " ) ", "Capability " + str(value))

                        console.print(f"[bold blue]Setting Menu: ({str(focusedSetting.id)})")
                        console.print(menu)
                        option = IntPrompt.ask(
                            prompt="Select an option from the menu above",
                            choices=[str(i) for i in range(len(settingValues) + option_offset)],
                            show_choices=False,
                        )
                        if option == 0:
                            focusedSetting = None
                        else:
                            console.print(
                                f"Changing setting {str(focusedSetting.id)} value to {str(settingValues[option - option_offset])}"
                            )

                            logger.debug(
                                f"Changing setting {str(focusedSetting.id)} to {str(settingValues[option - option_offset])}"
                            )
                            response = settingHandler.set(settingValues[option - option_offset])
                            logger.debug(response)
                            focusedSetting.value = settingValues[option - option_offset]
            else:
                presetDemo(cameraModel)
                collectionCacheInvalidated = True
                activeCacheInvalidated = True
                menuMode = MenuMode.MAIN


# Needed for poetry scripts defined in pyproject.toml
def entrypoint() -> None:
    main(parse_arguments())


if __name__ == "__main__":
    entrypoint()



# views.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed Aug 17 20:05:18 UTC 2022

"""GUI views and associated common functionality"""

# pylint: disable = arguments-differ

from __future__ import annotations
import enum
import json
import logging
import tkinter as tk
from tkinter import ttk, font
import tkinter.scrolledtext as ScrolledText
from abc import abstractmethod, ABC
from urllib.parse import urlparse, parse_qs
from typing import Any, Union, Sequence, Callable, Optional, Generator, Generic, TypeVar, cast, no_type_check

from PIL import Image, ImageTk, ImageDraw

from open_gopro.util import pretty_print
from open_gopro.demos.gui import models, THEME

MAX_TREEVIEW_ID = 1000000

GetterType = Callable[[], Any]
T = TypeVar("T")


class DefaultTextEntry(ttk.Entry, Generic[T]):
    """A Tkinter Entry with default text that disappears when clicked

    Args:
        default (T): default value to show in entry

    Raises:
        ValueError: _description_
    """

    def __init__(self, default: T, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        tk_var_type: type[tk.Variable]
        if isinstance(default, str):
            tk_var_type = tk.StringVar
        elif isinstance(default, bool):
            tk_var_type = tk.BooleanVar
        elif isinstance(default, int):
            tk_var_type = tk.IntVar
        elif isinstance(default, float):
            tk_var_type = tk.DoubleVar
        else:
            raise ValueError("entry_type must be one of [str, bool, int, float]")
        self.default: T = default
        self.var: tk.Variable = tk_var_type(value=self.default)
        self.insert(0, self.default)  # type: ignore
        self.bind("<FocusIn>", self.focus_in)

    def focus_in(self, *_: Any) -> None:
        """Called when entry is focused on, to remove default text

        Args:
            *_ (Any): Not used
        """
        if self.var.get() == self.default:
            self.delete(0, "end")

    def get(self) -> Optional[T]:  # type: ignore
        """Get the entry value. Will return none if the value is still default

        Returns:
            Optional[T]: value or None
        """
        return None if (current := cast(T, super().get())) == self.default else current


def popup_message(title: str, message: str, button: str = "OK") -> None:
    """Pop up a message in a separate window with an exit button.

    Args:
        title (str): Title of window
        message (str): message to display
        button (str): text to display in exit button. Defaults to "OK".
    """
    top = tk.Toplevel()
    # Get current mouse position
    abs_coord_x = top.winfo_pointerx() - top.winfo_rootx()
    abs_coord_y = top.winfo_pointery() - top.winfo_rooty()
    # Place window at mouse
    top.geometry(f"400x200+{abs_coord_x}+{abs_coord_y}")
    top.title(title)
    label = tk.Label(top, text=message, font=("Mistral 10 bold"), wraplength=350, pady=20)
    label.pack()

    def exit_btn() -> None:
        """On exit, close window"""
        top.destroy()
        top.update()

    btn = ttk.Button(top, text=button, command=exit_btn)
    btn.pack()


class View(ABC, tk.Widget):
    """View interface"""

    @no_type_check
    @abstractmethod
    def create_view(self) -> None:
        """Display the view (i.e. pack, grid, etc)"""
        raise NotImplementedError


class SplashScreen(View, tk.Frame):
    """Initial splash screen to connect a device"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        width = kwargs.pop("width", 200)
        height = kwargs.pop("height", 200)
        tk.Frame.__init__(self, *args, width=width, height=height, **kwargs)
        self.root = tk.Tk()
        self.style = ttk.Style(self.root)
        self.style.theme_use(THEME)
        self.root.title("Open GoPro Device Selection")
        self.defaultFont = font.nametofont("TkDefaultFont")
        self.root.geometry(f"{width}x{height}")
        # Beta labal
        self.beta_label = ttk.Label(
            self.root,
            text="This is beta software!! Not all commands have been tested yet.",
            font="26",
            foreground="red",
        )
        # Choose
        self.choose_label = ttk.Label(self.root, text="Choose either: ", font="18")
        # Specific device input
        self.device_select_input = ttk.Frame(self.root)
        self.device_select = DefaultTextEntry(default="GoPro Device Name", master=self.device_select_input)
        self.device_select.grid(row=0, column=0, sticky="NSEW", padx=5)
        self.device_select_button = ttk.Button(self.device_select_input, text="Connect")
        self.device_select_button.grid(row=0, column=1)
        # Or..
        self.or_label = ttk.Label(self.root, text="or...", font="18")
        # Auto connect
        self.auto_connect_button = ttk.Button(self.root, text="Auto Connect To First Found GoPro")

    def create_view(self, command: Callable) -> None:
        """Display the Splash Screen

        Args:
            command (Callable): command to be called when buttons are clicked
        """
        self.beta_label.pack(pady=10)
        self.choose_label.pack(pady=10)
        self.device_select_input.pack()
        self.device_select_button.configure(command=command)
        self.or_label.pack()
        self.auto_connect_button.pack(pady=10)
        self.auto_connect_button.configure(command=command)


class Menubar(View, tk.Menu):
    """Menu Bar view (i.e exit, etc)"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Menu.__init__(self, *args, **kwargs)
        self.master.configure(menu=self)  # type: ignore
        self.fileMenu = tk.Menu(self)

    def create_view(self) -> None:
        """Display the menu bar"""
        self.add_cascade(label="File", menu=self.fileMenu)


class StatusBar(View, tk.Frame):
    """Status Bar View"""

    LABEL_WIDTH = 15
    STATUS_WIDTH = 25

    class Color(enum.Enum):
        """Predefined color values"""

        RED = "#f5c6d0"
        YELLOW = "#fcffa1"
        GREEN = "#a1ffaa"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Frame.__init__(self, *args, **kwargs)
        self.statuses: list[ttk.Label] = []
        self.ble_label: ttk.Label
        self.ble_status: ttk.Label
        self.wifi_label: ttk.Label
        self.wifi_status: ttk.Label
        self.ready_label: ttk.Label
        self.ready_status: ttk.Label
        self.stream_label: ttk.Label
        self.stream_status: ttk.Label
        self.encoding_label: ttk.Label
        self.encoding_status: ttk.Label

    def create_view(self) -> None:
        """Display the status bar"""
        self.pack(side=tk.BOTTOM, fill=tk.X)
        # BLE status
        self.ble_label = ttk.Label(self, text="BLE State:", relief=tk.RAISED, width=self.LABEL_WIDTH)
        self.ble_label.grid(row=0, column=0)
        self.ble_status = ttk.Label(
            self, text="Idle", relief=tk.RAISED, width=self.STATUS_WIDTH, background=StatusBar.Color.RED.value
        )
        self.ble_status.grid(row=0, column=1)
        # Wifi Status
        self.wifi_label = ttk.Label(self, text="Wifi State", relief=tk.RAISED, width=self.LABEL_WIDTH)
        self.wifi_label.grid(row=0, column=2)
        self.wifi_status = ttk.Label(
            self, text="Idle", relief=tk.RAISED, width=self.STATUS_WIDTH, background=StatusBar.Color.RED.value
        )
        self.wifi_status.grid(row=0, column=3)
        # Busy Status
        self.ready_label = ttk.Label(self, text="Ready State:", relief=tk.RAISED, width=self.LABEL_WIDTH)
        self.ready_label.grid(row=0, column=4)
        self.ready_status = ttk.Label(self, text="", relief=tk.RAISED, width=self.STATUS_WIDTH)
        self.ready_status.grid(row=0, column=5)
        # Encoding status
        self.encoding_label = ttk.Label(self, text="Encoding State:", relief=tk.RAISED, width=self.LABEL_WIDTH)
        self.encoding_label.grid(row=0, column=6)
        self.encoding_status = ttk.Label(self, text="", relief=tk.RAISED, width=self.STATUS_WIDTH)
        self.encoding_status.grid(row=0, column=7)
        # Stream Status
        self.stream_label = ttk.Label(self, text="Stream State:", relief=tk.RAISED, width=self.LABEL_WIDTH)
        self.stream_label.grid(row=0, column=8)
        self.stream_status = ttk.Label(
            self, text="IDLE", relief=tk.RAISED, width=self.STATUS_WIDTH, background=StatusBar.Color.RED.value
        )
        self.stream_status.grid(row=0, column=9)

    @staticmethod
    def update_status(status: ttk.Label, color: StatusBar.Color, text: str) -> None:
        """Update a status with a given color and text

        Args:
            status (ttk.Label): label to update
            color (StatusBar.Color): color to set the label to
            text (str): text to display on the label
        """
        status.configure(text=text, background=color.value)


class Video(View, tk.Frame):
    """Displays a video source in a tkinter window"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Frame.__init__(self, *args, **kwargs)
        self.canvas = tk.Canvas(self)
        self.frame_label = ttk.Label(self, text="Video Stream")
        self.input_frame = tk.Frame(self)
        self.url_entry = DefaultTextEntry(default="Stream URL", master=self.input_frame)
        self.url_entry.grid(row=0, column=0, sticky="NSEW", padx=5)
        self.start_button = ttk.Button(self.input_frame, text="Start")
        self.start_button.grid(row=0, column=1)
        self.stop_button = ttk.Button(self.input_frame, text="Stop")
        self.stop_button.grid(row=0, column=2)
        self.input_frame.grid_columnconfigure(0, weight=6)
        self.input_frame.grid_columnconfigure(1, weight=1)
        self.input_frame.grid_columnconfigure(1, weight=1)

    def create_view(self, start: Callable, stop: Callable) -> None:
        """Display the initial video view with no image

        Args:
            start (Callable): command to call when start is clicked
            stop (Callable): command to call when stop is clicked
        """
        self.frame_label.pack(side=tk.TOP)
        self.input_frame.pack(side=tk.TOP, fill=tk.X)
        self.canvas.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
        self.pack(expand=True, fill=tk.BOTH)
        self.start_button.config(command=start)
        self.stop_button.config(command=stop)

    def display_frame(self, photo: ImageTk.PhotoImage) -> None:
        """Display an individual video frame

        Args:
            photo (ImageTk.PhotoImage): frame to display
        """
        self.canvas.create_image(0, 0, image=photo, anchor=tk.NW)


class Log(View, tk.Frame, ABC):
    """Log View interface"""

    class Format(enum.Enum):
        """Type of message to log

        For now, just sets the color
        """

        TX = "white"
        RX = "beige"
        ERROR = "#f5c6d0"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Frame.__init__(self, *args, **kwargs)

    @abstractmethod
    def create_log_entry(self, record: logging.LogRecord) -> None:
        """Display and individual log entry

        Args:
            record (logging.LogRecord): record to log
        """
        raise NotImplementedError


class TreeViewLog(Log):
    """A Log displayed as a TreeView"""

    # custom indicator images
    im_open = Image.new("RGBA", (15, 15), "#00000000")
    im_empty = Image.new("RGBA", (15, 15), "#00000000")
    draw = ImageDraw.Draw(im_open)
    draw.polygon([(0, 4), (14, 4), (7, 11)], fill="white", outline="black")
    im_close = im_open.rotate(90)

    class Message:
        """An individual log message to displayed by a Tree View Log

        Args:
            record (logging.LogRecord): logger record that will comprise this message
            fmt (str): string to format a tuple of values

        Raises:
            ValueError: Not able to build a Message from this log record
        """

        ASYNC_TOKEN = "-ASYNC-"

        # NOTE! Direction is from perspective of PC
        def __init__(self, record: logging.LogRecord, fmt: str) -> None:
            self.fmt = fmt
            self.message = record.message.strip("\n")
            if self.message[2] == "<":
                self.direction = Log.Format.TX
            elif self.message[-3] == ">":
                self.direction = Log.Format.RX
            else:
                raise ValueError("Log message must start with a directional arrow")
            self.time = record.asctime
            self.is_async = self.ASYNC_TOKEN in self.message
            self.sanitize_message()
            try:
                self.data: dict = json.loads("{" + self.message + "}")
            except Exception as e:
                raise ValueError(f"Can not build log message from {self.message}") from e
            self.data["target"] = self.data.pop("uuid", None) or self.data.pop("endpoint", None)
            self.sanitize_identifier()

        def sanitize_message(self) -> None:
            """Clean up the directional headers from the logger message string"""
            self.message = (
                self.message.replace(self.ASYNC_TOKEN, "").strip("<>").strip("-").strip("<>").replace("\t", "")
            )

        def sanitize_identifier(self) -> None:
            """Make the 'id' value human readable so that it looks the same for BLE and Wifi"""
            self.data["id"] = (
                self.data["id"]
                .replace("_", " ")
                .lower()
                .removeprefix("cmdid.")
                .removeprefix("querycmdid.")
                .removeprefix("actionid.")
                .removeprefix("gopro/")
                .replace("/", " ")
                .title()
            )

            try:
                self.data["id"] = self.data["id"].split("?")[0]
            except IndexError:
                pass
            self.data.pop("command", None)

            # Special case to handle wifi set setting response
            if self.data["id"].lower() == "camera setting":
                parsed_url = urlparse(self.data["target"])
                setting_id = int(parse_qs(parsed_url.query)["setting"][0])
                self.data["id"] = models.constants.SettingId(setting_id)

        @property
        def is_ok(self) -> bool:
            """Is the status ok or unknown?

            Returns:
                bool: True if yes, False otherwise
            """
            try:
                return self.data["status"].lower() in ["success", "unknown"]
            except KeyError:
                return True

        def logview_entries(self) -> Generator[tuple[str, tuple[str, ...]], None, None]:
            """From the Log Message, generate entries suitable for the Log View to consume

            Generates:
                1. top level description of message (formatted be self.fmt)
                2-N: individual elements of message as tuple(id, value)

            Yields:
                Generator[tuple[str, tuple[str, ...]], None, None]: see note in command description
            """
            work_dict = {**self.data, "log_time": self.time}
            # First yield top level
            out: list[str] = []
            for fmtstr in self.fmt:
                out.append(pretty_print(work_dict.pop(fmtstr, "")))
            yield pretty_print(work_dict.pop("id")), tuple(out)
            # Yield any additional items
            for key, value in work_dict.items():
                yield pretty_print(key), (pretty_print(str(value)),)

        def __str__(self) -> str:
            return json.dumps(self.data, indent=4)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.tv = ttk.Treeview(self)
        self.tv["columns"] = ("log_time", "protocol", "target", "status")
        self.tv.column("log_time", width=20)
        self.tv.column("protocol", width=20)
        self.tv.column("status", width=50)
        self.tv.column("target", width=50)
        self.tv.heading("log_time", text="Time")
        self.tv.heading("protocol", text="Protocol")
        self.tv.heading("status", text="Status")
        self.tv.heading("target", text="UUID/Endpoint")
        self.tv.tag_configure(Log.Format.TX.name, background=Log.Format.TX.value, foreground="black")
        self.tv.tag_configure(Log.Format.RX.name, background=Log.Format.RX.value, foreground="black")
        self.tv.tag_configure(Log.Format.ERROR.name, background=Log.Format.ERROR.value, foreground="black")
        self.sbv = tk.Scrollbar(self, orient=tk.VERTICAL)
        self.tv.config(yscrollcommand=self.sbv.set)
        self.sbv.config(command=self.tv.yview)
        self.index = 0
        self.img_open = ImageTk.PhotoImage(TreeViewLog.im_open, name="img_open", master=self.winfo_toplevel())
        self.img_close = ImageTk.PhotoImage(
            TreeViewLog.im_close, name="img_close", master=self.winfo_toplevel()
        )
        self.img_empty = ImageTk.PhotoImage(
            TreeViewLog.im_empty, name="img_empty", master=self.winfo_toplevel()
        )
        self.style = ttk.Style(self.winfo_toplevel())
        self.style.element_create(
            "Treeitem.myindicator",
            "image",
            "img_close",
            ("user1", "!user2", "img_open"),
            ("user2", "img_empty"),
            sticky="w",
            width=15,
        )
        # replace Treeitem.indicator by custom one
        self.style.layout(
            "Treeview.Item",
            [
                (
                    "Treeitem.padding",
                    {
                        "sticky": "nswe",
                        "children": [
                            ("Treeitem.myindicator", {"side": "left", "sticky": ""}),
                            ("Treeitem.image", {"side": "left", "sticky": ""}),
                            (
                                "Treeitem.focus",
                                {
                                    "side": "left",
                                    "sticky": "",
                                    "children": [("Treeitem.text", {"side": "left", "sticky": ""})],
                                },
                            ),
                        ],
                    },
                )
            ],
        )

    def create_view(self) -> None:
        """Display the TreeView Log"""
        self.sbv.pack(side=tk.RIGHT, fill=tk.Y)
        self.tv.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.pack(fill=tk.Y, expand=True)

    def create_log_entry_element(
        self, name: str, values: tuple[str, ...], fmt: Log.Format, parent: Optional[int] = None
    ) -> int:
        """Display an individual log entry element, potentially as a subtree

        Args:
            name (str): top level name of element
            values (tuple[str, ...]): values in element
            fmt (Log.Format): how to format the element
            parent (Optional[int], optional): Parent of element if part of subtree. Defaults to None.

        Returns:
            int: identifier of this created element
        """
        self.tv.insert(
            parent=str(parent or ""),
            index="end",
            iid=str(self.index),
            text=name,
            values=values,
            tags=fmt.name,
        )
        ret_index = self.index
        self.index += 1

        # Scroll to bottom
        self.tv.yview_moveto(1)

        return ret_index

    def create_log_entry(self, record: logging.LogRecord) -> None:
        """Create a log entry from a logging record

        Args:
            record (logging.LogRecord): record to analyze
        """
        try:
            msg = TreeViewLog.Message(record, self.tv["columns"])
        except ValueError:  # It doesn't have an arrow so we can't log it
            return
        logview_entry_gen = msg.logview_entries()
        name, values = next(logview_entry_gen)
        parent = self.create_log_entry_element(name, values, msg.direction if msg.is_ok else Log.Format.ERROR)
        for name, values in logview_entry_gen:
            self.create_log_entry_element(name, values, msg.direction, parent)


class TextLog(Log):
    """A Log View displayed as scrolled text"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.viewer = ScrolledText.ScrolledText(self, state="disabled")

    def create_view(self) -> None:
        """Display the text log view"""
        self.viewer.pack(expand=True, fill=tk.X)

    def create_log_entry(self, record: logging.LogRecord) -> None:
        """Create a text log entry from a logging record

        Args:
            record (logging.LogRecord): logging record to analyze
        """
        self.viewer.configure(state="normal")
        self.viewer.insert(tk.END, record.message + "\n")
        self.viewer.configure(state="disabled")
        # Autoscroll to the bottom
        self.viewer.yview(tk.END)


class ParamForm(View, tk.Frame):
    """A entry form to gather command arguments"""

    FRAME_TITLE_ROW = 0
    COMMAND_NAME_ROW = 1
    PARAM_START_ROW = 2

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Frame.__init__(self, *args, **kwargs)
        self.value_inputs: list[Union[tk.Variable, ttk.Entry]] = []
        self.arg_frames: list[tk.Frame] = []
        self.frame_label: ttk.Label
        self.command_label: ttk.Label
        self.value_label: ttk.Label
        self.value_menu: ttk.OptionMenu
        self.send_button: ttk.Button

    def create_view(self) -> None:
        """Display the parameter form"""
        self.frame_label = ttk.Label(self, text="Parameter Configuration", name="frame_label")
        self.frame_label.pack()
        self.pack(fill=tk.BOTH, expand=True, side=tk.TOP)

    @property
    def param_row(self) -> int:
        """Get the current parameter row as the argument entry form is built

        Returns:
            int: parameter row
        """
        return len(self.value_inputs) + self.PARAM_START_ROW

    @property
    def active_arg_frame(self) -> tk.Frame:
        """Get the active tkinter frame of the argument that is currently being built

        Returns:
            tk.Frame: active argument frame
        """
        return self.arg_frames[-1]

    def reset_view(self) -> None:
        """Destroy the current argument entries"""
        for widget in self.winfo_children():
            if "frame_label" not in str(widget):
                widget.destroy()
        self.value_inputs.clear()
        self.arg_frames.clear()

    def create_command(self, command: str) -> None:
        """Create a command label

        Args:
            command (str): name of command
        """
        self.command_label = ttk.Label(self, text=command)
        self.command_label.pack()

    def _create_with_args(self, param: str) -> None:
        """Create an entry that includes an argument

        Args:
            param (str): text of argument
        """
        self.arg_frames.append(tk.Frame(self))
        self.active_arg_frame.pack()
        self.value_label = ttk.Label(self.active_arg_frame, text=param, anchor="w")
        self.value_label.pack(side=tk.LEFT)

    def create_option_menu(
        self, param: str, options: Sequence[str], default: Optional[str] = None
    ) -> GetterType:
        """Create a drop down menu for an argument

        Args:
            param (str): name of argument
            options (Sequence[str]): all possible argument options
            default (Optional[str], optional): The default option. If none, the first option will be chosen.
                Defaults to None.

        Returns:
            GetterType: a getter to get this option menu value
        """
        self._create_with_args(param)
        value_input = tk.StringVar(self.active_arg_frame)
        default = default or options[0]
        value_input.set(default)
        self.value_menu = ttk.OptionMenu(self.active_arg_frame, value_input, default, *options)
        self.value_menu.pack(side=tk.RIGHT)
        self.value_inputs.append(value_input)
        return value_input.get

    def create_entry(self, param: str) -> GetterType:
        """Create a text entry for an argument

        Args:
            param (str): name of argument

        Returns:
            GetterType: A getter to get this text entry value
        """
        self._create_with_args(param)
        value_input = ttk.Entry(self.active_arg_frame)
        value_input.pack(side=tk.RIGHT)
        self.value_inputs.append(value_input)
        return value_input.get

    def create_button(self, label: str, command: Callable) -> None:
        """Create a button in the parameter entry form

        Args:
            label (str): string to place in button
            command (Callable): command to call when button is pressed
        """
        self.send_button = ttk.Button(self, text=label, command=command)
        self.send_button.pack()


class CommandPallette(View, tk.Frame):
    """A hierarchal display of commands by functionality for the user to select"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Frame.__init__(self, *args, **kwargs)
        self.tv = ttk.Treeview(self)
        self.sbv = tk.Scrollbar(self, orient=tk.VERTICAL)
        self.tv.config(yscrollcommand=self.sbv.set)
        self.sbv.config(command=self.tv.yview)
        self.frame_label = ttk.Label(self, text="Command Pallette")

    def create_view(self, commands: dict[str, list[str]]) -> None:
        """Display the command pallette

        Args:
            commands (dict[str, list[str]]): dictionary to be mapped to command pallette hierarchy
        """
        self.frame_label.pack()
        self.sbv.pack(side=tk.RIGHT, fill=tk.Y)
        self.tv.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.pack(fill=tk.Y, expand=True)
        # Fill up tree view
        parent_index = MAX_TREEVIEW_ID
        child_index = 0
        for parent, children in commands.items():
            self.tv.insert(parent="", index="end", iid=str(parent_index), text=parent)
            for command in children:
                self.tv.insert(parent=str(parent_index), index="end", iid=str(child_index), text=command)
                child_index += 1
            self.tv.item(str(parent_index), open=True)
            parent_index += 1


class StatusTab(View, tk.Frame):
    """A Status window to display current setting values / capabilities and status values

    Most recent updates will be highlighted
    """

    STATUS_ID_OFFSET = int(MAX_TREEVIEW_ID / 2)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        tk.Frame.__init__(self, *args, **kwargs)
        self.tv = ttk.Treeview(self)
        self.tv["columns"] = ("value", "capabilities")
        self.tv.column("value", width=200)
        self.tv.column("capabilities", width=400)
        self.tv.heading("value", text="Value")
        self.tv.heading("capabilities", text="Capabilities")
        self.tv.tag_configure("recent", background="#73ffff", foreground="black")

        self.sbv = tk.Scrollbar(self, orient=tk.VERTICAL)
        self.sbh = tk.Scrollbar(self, orient=tk.HORIZONTAL)
        self.tv.config(yscrollcommand=self.sbv.set)
        self.tv.config(xscrollcommand=self.sbh.set)
        self.sbv.config(command=self.tv.yview)
        self.sbh.config(command=self.tv.xview)
        self.setting_parent = str(MAX_TREEVIEW_ID)
        self.status_parent = str(int(self.setting_parent) + 1)
        self.setting_to_id: dict[int, str] = {}
        self.status_to_id: dict[int, str] = {}
        self.recent_tags: list[str] = []

    def create_view(self) -> None:
        """Display the Status Tab"""
        self.sbv.pack(side=tk.RIGHT, fill=tk.Y)
        self.sbh.pack(side=tk.TOP, fill=tk.X)
        self.tv.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.pack(fill=tk.Y, expand=True)
        # Fill up tree view. Start with just parents labels
        self.tv.insert(parent="", index="end", iid=self.setting_parent, text="Settings")
        self.tv.item(self.setting_parent, open=True)
        self.tv.insert(parent="", index="end", iid=self.status_parent, text="Statuses")
        self.tv.item(self.status_parent, open=True)

    def store_new_status(self, index: int) -> None:
        """Add a new status to the internal status index-id map

        Args:
            index (int): status index
        """
        self.status_to_id[index] = str(index + self.STATUS_ID_OFFSET)

    def store_new_setting(self, index: int) -> None:
        """Add a new setting to the internal setting index-id map

        Args:
            index (int): setting index
        """
        self.setting_to_id[index] = str(index)

    def _common_update(
        self,
        identifier: enum.IntEnum,
        update_map: dict[int, str],
        parent: str,
        store: Callable[[int], None],
        value: Optional[str] = None,
        capability: Optional[str] = None,
    ) -> None:
        """Common functionality to update a setting / status, storing new index if needed.

        One of value or capability must be non-None

        Args:
            identifier (enum.IntEnum): ID of setting / status to update
            update_map (dict[int, str]): index-to-id translation map
            parent (str): parent of this identifier
            store (Callable[[int], None]): method to store new index in update_map
            value (Optional[str], optional): value of update. Defaults to None.
            capability (Optional[str], optional): capability of update. Defaults to None.

        Raises:
            ValueError: _description_
        """
        # Empty string is a valid input so explicitly check if is None
        if value is None and capability is None:
            raise ValueError("Require value or capability")

        # Does it already exist?
        if tree_index := update_map.get(int(identifier)):
            # The values key in TreeView will return all of its value (values and caps from our perspective)
            current_values = self.tv.item(tree_index)["values"]
            # Need to check is not None because falsy empty string is a valid input
            new_values = tuple(
                new if new is not None else old for new, old in zip((value, capability), current_values)
            )
            self.tv.item(tree_index, tags=("recent",), values=new_values)
        else:
            store(int(identifier))
            tree_index = update_map[int(identifier)]
            self.tv.insert(
                parent=parent,
                index="end",
                iid=tree_index,
                text=pretty_print(identifier),
                values=(value or "", capability or ""),
                tags=("recent",),
            )
        self.recent_tags.append(tree_index)

    def update_setting(self, identifier: enum.IntEnum, value: str) -> None:
        """Display a setting value update

        Args:
            identifier (enum.IntEnum): setting that has been updated
            value (str): new value of setting
        """
        self._common_update(
            identifier,
            update_map=self.setting_to_id,
            store=self.store_new_setting,
            parent=self.setting_parent,
            value=value,
        )

    def update_capability(self, identifier: enum.IntEnum, capability: str) -> None:
        """Display a setting capability update

        Args:
            identifier (enum.IntEnum): setting that has been updated
            capability (str): new capability of setting
        """
        self._common_update(
            identifier,
            update_map=self.setting_to_id,
            parent=self.setting_parent,
            store=self.store_new_setting,
            capability=capability,
        )

    def update_status(self, identifier: enum.IntEnum, value: str) -> None:
        """Display a status value update

        Args:
            identifier (enum.IntEnum): status that has been updated
            value (str): new value of status
        """
        self._common_update(
            identifier,
            update_map=self.status_to_id,
            parent=self.status_parent,
            store=self.store_new_status,
            value=value,
        )

    def clear_recent_tags(self) -> None:
        """Set all currently marked 'recent' updates as not 'recent'"""
        for tree_index in self.recent_tags:
            self.tv.item(tree_index, tags=())
        self.recent_tags.clear()



# photo.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:45 PM

"""Entrypoint for taking a picture demo."""

import argparse
from pathlib import Path
from typing import Optional

from rich.console import Console

from open_gopro import GoPro, Params
from open_gopro.util import setup_logging, add_cli_args_and_parse

console = Console()  # rich consoler printer


def main(args: argparse.Namespace) -> None:
    logger = setup_logging(__name__, args.log)

    def exception_cb(exception: Exception) -> None:  # pylint: disable=unused-variable
        logger.error(f"IN MAIN ==> {exception}")

    gopro: Optional[GoPro] = None
    try:
        with GoPro(args.identifier, wifi_interface=args.wifi_interface, exception_cb=exception_cb) as gopro:
            # Configure settings to prepare for photo
            if gopro.is_encoding:
                gopro.ble_command.set_shutter(Params.Toggle.DISABLE)
            gopro.ble_setting.video_performance_mode.set(Params.PerformanceMode.MAX_PERFORMANCE)
            gopro.ble_setting.max_lens_mode.set(Params.MaxLensMode.DEFAULT)
            gopro.ble_setting.camera_ux_mode.set(Params.CameraUxMode.PRO)
            gopro.ble_command.set_turbo_mode(False)
            assert gopro.ble_command.load_preset_group(Params.PresetGroup.PHOTO).is_ok

            # Get the media list before
            media_set_before = set(x["n"] for x in gopro.wifi_command.get_media_list().flatten)
            # Take a photo
            console.print("Capturing a photo...")
            assert gopro.ble_command.set_shutter(Params.Toggle.ENABLE).is_ok

            # Get the media list after
            media_set_after = set(x["n"] for x in gopro.wifi_command.get_media_list().flatten)
            # The photo (is most likely) the difference between the two sets
            photo = media_set_after.difference(media_set_before).pop()
            # Download the photo
            console.print("Downloading the photo...")
            gopro.wifi_command.download_file(camera_file=photo, local_file=args.output)
            console.print(f"Success!! :smiley: File has been downloaded to {args.output}")

    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Shutting down...")

    if gopro:
        gopro.close()
    console.print("Exiting...")


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Connect to a GoPro camera, take a photo, then download it.")
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Where to write the photo to. If not set, write to 'photo.jpg'",
        default=Path("photo.jpg"),
    )
    return add_cli_args_and_parse(parser)


# Needed for poetry scripts defined in pyproject.toml
def entrypoint() -> None:
    main(parse_arguments())


if __name__ == "__main__":
    entrypoint()



# responses.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:49 PM

"""Any responses that are returned from GoPro commands."""

from __future__ import annotations
from abc import abstractmethod, ABC
import enum
import logging
from collections import defaultdict
from typing import (
    Any,
    Optional,
    Union,
    Iterator,
    ItemsView,
    ValuesView,
    KeysView,
    Final,
    ClassVar,
    Protocol,
    TypeVar,
    Generic,
    Callable,
)

import requests

from open_gopro.exceptions import ResponseParseError
from open_gopro.constants import (
    ActionId,
    FeatureId,
    GoProEnum,
    StatusId,
    CmdId,
    ErrorCode,
    SettingId,
    QueryCmdId,
    ResponseType,
    CmdType,
    GoProUUIDs,
)
from open_gopro.util import scrub, jsonify
from open_gopro.ble import BleUUID
from open_gopro.proto.response_generic_pb2 import EnumResultGeneric

logger = logging.getLogger(__name__)

CONT_MASK: Final = 0b10000000
HDR_MASK: Final = 0b01100000
GEN_LEN_MASK: Final = 0b00011111
EXT_13_BYTE0_MASK: Final = 0b00011111

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")


class Parser(ABC, Generic[T, T_co]):
    """Base Parser Interface

    Supports addition to other Parser's
    """

    def __init__(self) -> None:
        self._parsers: list[Callable] = [self.parse]

    def _combined_parse(self, data: T) -> T_co:
        """Parse data using all parsers serially

        Args:
            data (T): input data to parse

        Returns:
            T_co: parsed output
        """
        response = self._parsers[0](data)
        for parser in self._parsers[1:]:
            response = parser(response)
        return response

    @abstractmethod
    def parse(self, data: T) -> T_co:  # pylint: disable=method-hidden
        """Initial (and potentially only) parser method

        Args:
            data (T): input data to parse



        Returns:
            T_co: parsed output
        """
        raise NotImplementedError

    def __add__(self, other: Parser) -> Parser:
        if not isinstance(other, Parser):
            raise TypeError("Can only add Parser's")
        self._parsers.append(other.parse)
        self.parse = self._combined_parse  # type: ignore
        return self

    def __radd__(self, other: Parser) -> Parser:
        return other.__add__(self)


class BytesParser(Parser[bytes, T_co]):
    """Bytes parser protocol to be used by non-construct parsers"""

    @abstractmethod
    def parse(self, data: bytes) -> T_co:
        """Parse input bytes to output

        Args:
            data (bytes): data to parsed

        Returns:
            T_co: parsed output
        """
        raise NotImplementedError


class BytesBuilder(Protocol):
    """Base bytes serializer protocol definition"""

    def build(self, obj: Any) -> bytes:
        """Build bytestream from object

        # noqa: DAR202

        Args:
            obj (Any): object to serialize

        Returns:
            bytes: serialized bytestream
        """


class BytesParserBuilder(BytesParser[T_co], BytesBuilder):
    """Class capable of both building / parsing bytes to / from object"""

    @abstractmethod
    def parse(self, data: bytes) -> T_co:
        """Parse input bytes to output

        Args:
            data (bytes): data to parsed

        Returns:
            T_co: parsed output
        """
        raise NotImplementedError

    def build(self, obj: Any) -> bytes:
        """Build bytestream from object

        # noqa: DAR202

        Args:
            obj (Any): object to serialize

        Returns:
            bytes: serialized bytestream
        """


class JsonParser(Parser[dict, dict]):
    """The JSON Parser interface"""

    @abstractmethod
    def parse(self, data: dict) -> dict:
        """Parse input json to modified output json

        Args:
            data (dict): data to parsed

        Returns:
            dict: parsed output
        """
        raise NotImplementedError


class Header(enum.Enum):
    """Packet Headers."""

    GENERAL = 0b00
    EXT_13 = 0b01
    EXT_16 = 0b10
    RESERVED = 0b11
    CONT = enum.auto()


class GoProResp:
    """A flexible object to be used to encapsulate all GoPro responses.

    It can be instantiated with varying levels of information and filled out as more is received.
    The end user should never need to create a response and will only be consuming them.

    It is mostly a wrapper around a JSON-like dictionary (GoProResp.data)
    For example, first send the command and store the response:

    >>> response = ble_setting.resolution.get_value()

    Now let's inspect the responses various attributes / properties:

    >>> print(response.status)
    ErrorCode.SUCCESS
    >>> print(response.is_ok)
    True
    >>> print(response.id)
    QueryCmdId.GET_SETTING_VAL
    >>> print(response.cmd)
    QueryCmdId.GET_SETTING_VAL
    >>> print(response.uuid)
    GoProUUIDs.CQ_QUERY_RESP

    Now let's print it as (as JSON):

    >>> print(response.data)
    {
        "status": "SUCCESS",
        "id": "QueryCmdId.GET_SETTING_VAL",
        "SettingId.RESOLUTION": [
            "RES_1080"
        ]
    }
    """

    _feature_action_id_map: ClassVar[dict[FeatureId, list[ActionId]]] = defaultdict(list)
    _global_parsers: ClassVar[dict[ResponseType, Parser]] = {}

    class _State(enum.Enum):
        """Describes the state of the GoProResp."""

        INITIALIZED = enum.auto()
        ACCUMULATED = enum.auto()
        PARSED = enum.auto()
        ERROR = enum.auto()

    class Protocol(enum.Enum):
        """Protocol that Command will be sent on."""

        BLE = "BLE"
        WIFI = "WIFI"

    def __init__(
        self,
        meta: list[ResponseType],
        parser: Optional[Parser] = None,
        status: ErrorCode = ErrorCode.SUCCESS,
        raw_packet: Optional[Union[bytearray, dict[str, Any]]] = None,
    ) -> None:
        """Constructor

        Args:
            meta (list[ResponseType]): A list of all information known about the response.
            parser (Optional[Parser]): Optional parser. If not passed, parser will be found from global
                parsers
            status (ErrorCode): A status if known at time of instantiation. Defaults to ErrorCode.SUCCESS.
            raw_packet (Optional[Union[bytearray, dict[str, Any]]]): The unparsed input if known at time of
                instantiation. Defaults to None.
        """
        # A list describing all of the currently known information about the response.
        # This will be appended to as more information is discovered. The various properties of GoProResp will
        # use this list to parse out their relevant information.
        self._meta: list[ResponseType] = meta
        # Parsers to use to parse this response
        self._parser = parser

        self.status: ErrorCode = status
        """Status of the response"""

        # Start with empty list as default value in case we need to append.
        # If we end up not needing a list, we will just overwrite the default
        self.data: dict[Any, Any] = defaultdict(list)
        """Response data which is really JSON data stored as a dict"""

        self._raw_packet = raw_packet
        self._bytes_remaining = 0
        self._state: GoProResp._State = GoProResp._State.INITIALIZED

    @classmethod
    def _from_read_response(cls, uuid: BleUUID, data: bytearray) -> GoProResp:
        """Build a GoProResp from a read response.

        Args:
            uuid (BleUUID): BleUUID that read command was received from
            data (bytearray): received bytestream

        Returns:
            GoProResp: created instance
        """
        resp = cls(meta=[uuid], status=ErrorCode.SUCCESS, raw_packet=data)
        resp._parse()
        return resp

    @classmethod
    def _from_http_response(
        cls, parser: Optional[JsonParser], response: requests.models.Response
    ) -> GoProResp:
        """Build a GoProResp from an HTTP response from the requests package.

        Args:
            parser (Optional[JsonParser]): parsers to use to parse received data
            response (requests.models.Response): HTTP response

        Returns:
            GoProResp: created instance
        """
        resp = cls(
            meta=[response.url],
            parser=parser,
            status=ErrorCode.SUCCESS if response.ok else ErrorCode.ERROR,
            raw_packet=response.json() if response.text else {},
        )
        resp._parse()
        return resp

    @classmethod
    def _get_response_meta(cls, data: bytes, uuid: BleUUID) -> list[ResponseType]:
        """Get a response's meta information from raw bytes and the UUID it was received on

        Args:
            data (bytes): bytes received as BLE notification
            uuid (BleUUID): UUID response was received on

        Returns:
            list[ResponseType]: List of meta information in order from least to most specific
        """
        meta: list[ResponseType] = [uuid]

        # If it's a protobuf command
        identifier = data[0]
        try:
            FeatureId(identifier)
            meta.append(ActionId(data[1]))
        # Otherwise it's a TLV command
        except ValueError:
            if uuid is GoProUUIDs.CQ_SETTINGS_RESP:
                meta.append(SettingId(identifier))
            elif uuid is GoProUUIDs.CQ_QUERY_RESP:
                meta.append(QueryCmdId(identifier))
            elif uuid in [GoProUUIDs.CQ_COMMAND_RESP, GoProUUIDs.CN_NET_MGMT_RESP]:
                meta.append(CmdId(identifier))
            # Else case is direct so the UUID (which is already in meta) is the identifier
        return meta

    @classmethod
    def _add_feature_action_id_mapping(cls, feature_id: FeatureId, action_id: ActionId) -> None:
        """Add a feature id-to-action id mapping entry

        Args:
            feature_id (FeatureId): Feature ID of protobuf command
            action_id (ActionId): Action ID of protobuf command
        """
        cls._feature_action_id_map[feature_id].append(action_id)

    @classmethod
    def _add_global_parser(cls, identifier: ResponseType, parser: Parser) -> None:
        """Add a global parser that can be accessed by this class's class methods

        Args:
            identifier (ResponseType): identifier to add parser for
            parser (Parser): parser to add
        """
        cls._global_parsers[identifier] = parser

    @classmethod
    def _get_query_container(cls, identifier: Union[SettingId, StatusId]) -> Optional[Callable]:
        """Attempt to get a callable that will translate an input value to the ID-appropriate value.

        For example, _get_query_container(SettingId.RESOLUTION) will return
        :py:class:`open_gopro.api.params.Resolution`

        As another example, _get_query_container(StatusId.TURBO_MODE) will return bool()

        Note! Not all ID's are currently parsed so None will be returned if the container does not exist

        Args:
            identifier (Union[SettingId, StatusId]): identifier to find container for

        Returns:
            Callable: container if found else None
        """
        try:
            return lambda x, pb=cls._global_parsers[identifier]: pb.parse(pb.build(x))
        except KeyError:
            return None

    @classmethod
    def _get_global_parser(cls, identifier: ResponseType) -> Optional[Parser]:
        """Get a globally defined parser for the given ID.

        Currently, only BLE uses globally defined parsers

        Args:
            identifier (ResponseType): ID to get parser for

        Returns:
            Optional[Parser]: parser if found, else None
        """
        return cls._global_parsers.get(identifier)

    def __eq__(self, obj: object) -> bool:
        if isinstance(obj, GoProEnum):
            return self.identifier == obj
        if isinstance(obj, GoProResp):
            return self.identifier == obj.identifier
        raise TypeError("Equal can only compare GoProResp and ResponseType")

    def __getitem__(self, key: Any) -> Any:
        return self.data[key]

    def __contains__(self, item: Any) -> bool:
        return item in self.data

    def __iter__(self) -> Iterator:
        return iter(self.data)

    def __str__(self) -> str:
        return jsonify(self._as_dict())

    def _as_dict(self) -> dict[str, Any]:
        """Return the response as a dict

        Returns:
            dict[str, Any]: response as dict
        """
        work_dict = dict(id=self.identifier, protocol=self.protocol.name, status=self.status.name)
        if self.cmd:
            work_dict["command"] = self.cmd
        if self.uuid:
            work_dict["uuid"] = self.uuid
        if self.endpoint:
            work_dict["endpoint"] = self.endpoint
        return {**work_dict, **self.data}

    def __repr__(self) -> str:
        return f"GoProResp <{str(self.identifier)}: {self._state}>"

    def items(self) -> ItemsView[Any, Any]:
        """Pass-through to access data dict's "items" method

        Returns:
            ItemsView[Any, Any]: all of data's keys
        """
        return self.data.items()

    def keys(self) -> KeysView[Any]:
        """Pass-through to access data dict's "keys" method

        Returns:
            ItemsView[Any, Any]: all of data's items
        """
        return self.data.keys()

    def values(self) -> ValuesView[Any]:
        """Pass-through to access data dict's "values" method

        Returns:
            ItemsView[Any, Any]: all of data's values
        """
        return self.data.values()

    @property
    def protocol(self) -> GoProResp.Protocol:
        """Get the protocol that the response was received on

        Returns:
            GoProResp.Protocol: protocol
        """
        return GoProResp.Protocol.BLE if self.uuid else GoProResp.Protocol.WIFI

    @property
    def is_received(self) -> bool:
        """Has the response been completely received?

        Returns:
            bool: True if completely received, False if not
        """
        return self._state is not GoProResp._State.INITIALIZED

    @property
    def is_parsed(self) -> bool:
        """Has the response been successfully parsed?

        Returns:
            bool: True if it has been parsed, False if not
        """
        return self._state is GoProResp._State.PARSED

    @property
    def is_ok(self) -> bool:
        """Are there any errors in this response?

        Returns:
            bool: True if the response is ok (i.e. there are no errors), False otherwise
        """
        return self.status in [ErrorCode.SUCCESS, ErrorCode.UNKNOWN]

    @property
    def identifier(self) -> ResponseType:
        """Get the identifier of the response.

        Will vary depending on what type of response this is:

        - for a direct BLE read / write to a characteristic, it will be a :py:class:`open_gopro.ble.services.BleUUID`
        - for a BLE command response, it will be a :py:class:`open_gopro.constants.CmdType`
        - for a BLE setting / status response / update, it will be a :py:class:`open_gopro.constants.SettingId`
            or :py:class:`open_gopro.constants.StatusId`
        - for an HTTP response, it will be a string of the HTTP endpoint

        Returns:
            ResponseType: the identifier
        """
        return self._meta[-1]

    @property
    def cmd(self) -> Optional[CmdType]:
        """Attempt to get the command ID of the response.

        If the response is not a command response, it won't have a command.

        Returns:
            Optional[CmdType]: Command ID if relevant, otherwise None
        """
        for x in self._meta:
            if isinstance(x, (QueryCmdId, CmdId)):
                return x
        return None

    @property
    def uuid(self) -> Optional[BleUUID]:
        """Attempt to get the BleUUID the response was received on.

        If the response is not a BLE response, it won't have a BleUUID.

        Returns:
            Optional[BleUUID]: BleUUID if relevant, otherwise None
        """
        for x in self._meta:
            if isinstance(x, BleUUID):
                return x
        return None

    @property
    def endpoint(self) -> Optional[str]:
        """Attempt to get the endpoint that response was received from.

        If the response is not an HTTP response, it won't have an endpoint.

        Returns:
            Optional[str]: Endpoint if relevant, otherwise None
        """
        for x in self._meta:
            if isinstance(x, str):
                return x
        return None

    @property
    def is_protobuf(self) -> bool:
        """Is this response a protobuf response

        Returns:
            bool: Yes if true, No otherwise
        """
        return bool({ActionId, FeatureId}.intersection({type(x) for x in self._meta}))

    @property
    def is_direct_read(self) -> bool:
        """Is this response a direct read of a BLE characteristic

        Returns:
            bool: Yes if true, No otherwise
        """
        return len(self._meta) == 1 and isinstance(self.identifier, BleUUID)

    @property
    def flatten(self) -> Any:
        """Attempt to flatten / simplify the JSON response (GoProResp.data).

        - If there is only one entry in the JSON dict and it is a single value, return the value
        - If there is only one entry in the JSON dict and it is a list of values, return the list
        - Otherwise, just return the JSON dict

        Returns:
            Any: A single value, a list of values, or the JSON dict
        """
        values = list(self.data.values())
        if len(values) == 1 and type(values[0] not in [list, dict]):
            return values[0]

        if len(values) == 1 and type(values[0] is list):
            return values

        return self.data

    # to be @overload'ed if anything besides BLE notifications need to be accumulated
    def _accumulate(self, data: bytes) -> None:
        """Accumulate BLE byte data.

        If there is no more data left to accumulate, the bytestream will be parsed.

        Args:
            data (bytes): byte level BLE data
        """
        buf = bytearray(data)
        if buf[0] & CONT_MASK:
            buf.pop(0)
        else:
            # This is a new packet so start with an empty byte array
            self._raw_packet = bytearray([])
            hdr = Header((buf[0] & HDR_MASK) >> 5)
            if hdr is Header.GENERAL:
                self._bytes_remaining = buf[0] & GEN_LEN_MASK
                buf = buf[1:]
            elif hdr is Header.EXT_13:
                self._bytes_remaining = ((buf[0] & EXT_13_BYTE0_MASK) << 8) + buf[1]
                buf = buf[2:]
            elif hdr is Header.EXT_16:
                self._bytes_remaining = (buf[1] << 8) + buf[2]
                buf = buf[3:]

        # Append payload to buffer and update remaining / complete
        assert isinstance(self._raw_packet, bytearray)
        self._raw_packet.extend(buf)
        self._bytes_remaining -= len(buf)

        if self._bytes_remaining < 0:
            logger.error("received too much data. parsing is in unknown state")
        elif self._bytes_remaining == 0:
            self._state = GoProResp._State.ACCUMULATED

    def _parse(self) -> None:
        """Parse the accumulated response (either from a BLE bytestream or an HTTP JSON dict).

        Raises:
            NotImplementedError: Parsing for this id is not yet supported
            ResponseParseError: Error when parsing data
            RuntimeError: unexpected state
        """
        # Is this a BLE response?
        if isinstance(self._raw_packet, bytearray):
            assert self.uuid
            buf: bytearray = self._raw_packet
            self._meta = self._get_response_meta(buf, self.uuid)

            if not self.is_direct_read:  # length byte
                buf.pop(0)
            if self.is_protobuf:  # feature ID byte
                buf.pop(0)

            try:
                query_type: Optional[Union[type[StatusId], type[SettingId], StatusId, SettingId]] = None
                # Need to delineate QueryCmd responses between settings and status
                if not self.is_protobuf:
                    if isinstance(self.identifier, (SettingId, StatusId)):
                        query_type = self.identifier
                    elif isinstance(self.identifier, QueryCmdId):
                        if self.identifier in [
                            QueryCmdId.GET_STATUS_VAL,
                            QueryCmdId.REG_STATUS_VAL_UPDATE,
                            QueryCmdId.UNREG_STATUS_VAL_UPDATE,
                            QueryCmdId.STATUS_VAL_PUSH,
                        ]:
                            query_type = StatusId
                        elif self.identifier is QueryCmdId.GET_SETTING_NAME:
                            raise NotImplementedError
                        else:
                            query_type = SettingId

                # Query (setting get value, status get value, etc.)
                if query_type:
                    self.status = ErrorCode(buf[0])
                    buf = buf[1:]
                    # Parse all parameters
                    while len(buf) != 0:
                        param_id = query_type(buf[0])  # type: ignore
                        param_len = buf[1]
                        buf = buf[2:]
                        # Special case where we register for a push notification for something that does not yet
                        # have a value
                        if param_len == 0:
                            self.data[param_id] = []
                            continue
                        param_val = buf[:param_len]
                        buf = buf[param_len:]

                        # Add parsed value to response's data dict
                        try:
                            if not (parser := self._get_global_parser(param_id)):
                                # We don't have defined params for all ID's yet. Just store raw bytes
                                logger.warning(f"No parser defined for {param_id}")
                                self.data[param_id] = param_val
                                continue
                            # These can be more than 1 value so use a list
                            if self.cmd in [
                                QueryCmdId.GET_CAPABILITIES_VAL,
                                QueryCmdId.REG_CAPABILITIES_UPDATE,
                                QueryCmdId.SETTING_CAPABILITY_PUSH,
                            ]:
                                # Parse using parser from global map and append
                                self.data[param_id].append(parser.parse(param_val))
                            else:
                                # Parse using parser from map and set
                                self.data[param_id] = parser.parse(param_val)
                        except ValueError:
                            # This is the case where we receive a value that is not defined in our params.
                            # This shouldn't happen and means the documentation needs to be updated. However, it
                            # isn't functionally critical
                            logger.warning(f"{param_id} does not contain a value {param_val}")
                            self.data[param_id] = param_val
                else:  # Commands,  Protobuf, and direct Reads
                    if is_cmd := isinstance(self.identifier, CmdId):
                        # All (non-protobuf) commands have a status
                        self.status = ErrorCode(buf[0])
                        buf = buf[1:]
                    # Use parser if explicitly passed otherwise get global parser
                    if not (parser := self._get_global_parser(self.identifier) or self._parser) and not is_cmd:
                        error_msg = f"No parser exists for {self.identifier}"
                        logger.error(error_msg)
                        raise ResponseParseError(str(self.identifier), self._raw_packet, msg=error_msg)
                    # Parse payload if a parser was found.
                    if parser:
                        self.data = parser.parse(buf)
                    # Attempt to determine and / or extract status (we already got command status above)
                    if self.is_direct_read and len(self._raw_packet):
                        # Assume success on direct reads if there was any data
                        self.status = ErrorCode.SUCCESS
                    elif (
                        self.is_protobuf and self.data and "result" in self.data
                    ):  # Check for result field in protobuf's
                        self.status = (
                            ErrorCode.SUCCESS
                            if self.data["result"] == EnumResultGeneric.RESULT_SUCCESS
                            else ErrorCode.ERROR
                        )
                    elif not is_cmd:
                        self.status = ErrorCode.UNKNOWN
            except KeyError as e:
                self._state = GoProResp._State.ERROR
                raise ResponseParseError(str(self.identifier), buf) from e

        # Is this an HTTP response?
        elif isinstance(self._raw_packet, dict):
            # Is there a parser for this? Most of them do not have one yet.
            self.data = self._parser.parse(self._raw_packet) if self._parser else self._raw_packet

        # This should never happen
        else:
            raise RuntimeError(
                f"Unexpected data type ({str(type(self._raw_packet))}) when attempting to parse response"
            )

        # Recursively scrub away parsing artifacts
        scrub(self.data, "_io")
        scrub(self.data, "status")
        self._state = GoProResp._State.PARSED



# models.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed Aug 17 20:05:18 UTC 2022

"""GUI models and associated common functionality"""

# pylint: disable = reimported, unused-import
# NOTE! The reason for the seemingly redundant, unnecessary import here is because we are using eval
# to dynamically  build commands

from __future__ import annotations
import re
import enum
import time
import logging
import inspect
from pathlib import Path
import datetime
import typing
from typing import Pattern, Any, Callable, Generator, Optional, no_type_check, Union, Final

import construct

from open_gopro import GoPro, constants
from open_gopro.api import BleStatus, WifiSetting, BleSetting
from open_gopro.interface import BleCommand, WifiCommand, Commands, Command
import open_gopro.api.params
import open_gopro.api.params as Params
from open_gopro.responses import GoProResp, ResponseType

PREVIEW_STREAM_URL: Final = r"udp://127.0.0.1:8554"

logger = logging.getLogger(__name__)


class CompoundGoPro(GoPro):
    """A GoPro that supports sending compound commands"""

    def __init__(self, target: Optional[Pattern] = None) -> None:
        """Constructor

        Args:
            target (Optional[Pattern], optional): BLE device (camera) to search for. Defaults to None (first
                found camera will be connected to).
        """

        super().__init__(target)
        self.compound_command = CompoundCommands(self)


class GoProModel:
    """GoPro model interface for controllers"""

    class Update(enum.Enum):
        """The type of update that the response / element corresponds to"""

        SETTING = enum.auto()
        STATUS = enum.auto()
        CAPABILITY = enum.auto()
        PROTOBUF = enum.auto()

    def __init__(self) -> None:
        # Initial instantiation just to get command strings
        self.gopro: CompoundGoPro = CompoundGoPro()

    def start(self, identifier: Optional[Pattern]) -> None:
        """Open the model (i.e. connect BLE and Wifi to camera)

        Args:
            identifier (Optional[Pattern]): regex to connect to. Defaults to None (connect to first camera)
        """
        # Reinstantiate
        self.gopro = CompoundGoPro(target=identifier)
        self.gopro.open()

    # NOTE: the following properties must be evaluated dynamically since self.gopro is changing

    @property
    def _command_types(self) -> list[Commands]:
        """Get the top level containers of the command types supported by the GoPro model

        Returns:
            list[Commands]: list of command type containers
        """
        return [
            self.gopro.ble_command,
            self.gopro.ble_setting,
            self.gopro.ble_status,
            self.gopro.wifi_command,
            self.gopro.wifi_setting,
            self.gopro.compound_command,
        ]

    @property
    def commands(self) -> list[Command]:
        """Get all of the available BLE and Wifi Commands

        Returns:
            list[Command]: list of available commands
        """
        commands = []
        for command_type in self._command_types:
            c = list(command_type)
            c.sort(key=lambda x: str(x))
            commands.extend(c)
        return commands

    @property
    def command_dict(self) -> dict[str, list[str]]:
        """Get flattened dictionary of commands indexed by their string name

        Returns:
            dict[str, list[str]]: flattened dict
        """
        d = {}
        for commands in self._command_types:
            d[type(commands).__name__] = [str(command) for command in commands]
            d[type(commands).__name__].sort()
        return d

    @classmethod
    def is_ble(cls, command: Command) -> bool:
        """Is this command a BLE command?

        Args:
            command (Command): command to analyze

        Returns:
            bool: True if yes, False otherwise
        """
        return type(command) in (BleSetting, BleStatus, BleCommand)

    @classmethod
    def is_wifi(cls, command: Command) -> bool:
        """Is this command a Wifi Command?

        Args:
            command (Command): command to analyze

        Returns:
            bool: True if yes, False otherwise
        """
        return type(command) in (WifiCommand, WifiSetting)

    @classmethod
    def is_setting(cls, command: Command) -> bool:
        """Is this command a setting?

        Args:
            command (Command): command to analyze

        Returns:
            bool: True if yes, False otherwise
        """
        return type(command) in (BleSetting, WifiSetting)

    @classmethod
    def is_status(cls, command: Command) -> bool:
        """Is this command a status?

        Args:
            command (Command): command to analyze

        Returns:
            bool: True if yes, False otherwise
        """
        return isinstance(command, BleStatus)

    @classmethod
    def is_command(cls, command: Command) -> bool:
        """Is this command a command (i.e. not setting or status)?

        Args:
            command (Command): command to analyze

        Returns:
            bool: True if yes, False otherwise
        """
        return (
            isinstance(command, (BleCommand, WifiCommand, CompoundCommand))
            and not cls.is_status(command)
            and not cls.is_setting(command)
        )

    @classmethod
    @no_type_check
    def get_args_info(cls, command: Command) -> tuple[list[str], list[type]]:
        """Get the argument names and types for a given command

        Args:
            command (Command): command to analyze

        Returns:
            tuple[list[str], list[type]]: (list[argument names], list[argument types])
        """
        arg_types: list[type] = []
        arg_names: list[str] = []
        method_info = inspect.getfullargspec(command if cls.is_command(command) else command.set)
        for arg in method_info.args[1:]:
            try:
                # Assume this is a generic and try to get the generic type of its original class
                arg_type = re.search(r"\[.*\]", str(command.__orig_class__))[0].strip("[]")
            except AttributeError:
                # This is not a generic so use the annotations from inspect
                arg_type = method_info.annotations[arg]

            # TODO temporary workaround to avoid handling parameter sequences. This is not an actual solution
            if "optional" in arg_type.lower():
                continue
            arg_types.append(eval(arg_type))  # pylint: disable = eval-used
            arg_names.append(arg)
        return arg_names, arg_types

    def get_command_info(
        self, command: Command
    ) -> tuple[list[Callable], list[Callable], list[type], list[str]]:
        """For a given command, get its adapters, validator, argument types, and argument names

        Args:
            command (Command): command to analyze

        Raises:
            Exception: unhandled type of argument

        Returns:
            tuple[list[Callable], list[Callable], list[type], list[str]]:
                (list[adapters], list[validators], list[arg_types], list[arg_names])
        """
        adapters: list[Callable] = []
        validators: list[Callable] = []
        names, arg_types = self.get_args_info(command)

        # Build adapters and validators
        # NOTE! These lambdas must define default variables since they are lost once the for loop scope exits
        for arg_type in arg_types:
            if "enum" in str(arg_type).lower():
                try:
                    format_field = command.param_builder.fmtstr  # type: ignore
                    for construct_field in (construct.Int8ub, construct.Int32ub, construct.Int16ub):
                        if construct_field.fmtstr == format_field:
                            validators.append(lambda x, cs=construct_field: cs.build(int(x)))
                            adapters.append(lambda x, adapt=arg_type: adapt[x])
                            break
                    else:
                        raise Exception("unrecognized format field" + str(format_field))
                except AttributeError:
                    validators.append(lambda x, adapt=arg_type: adapt(x))
                    adapters.append(lambda x, adapt=arg_type: adapt[x])
            elif arg_type is datetime.datetime:
                # Special case to be handled by controller
                adapters.append(datetime.datetime)
                validators.append(datetime.datetime)
            else:
                validators.append(lambda _: True)
                if arg_type is bool:
                    adapters.append(lambda x: bool(x.lower() == "true"))
                if arg_type in [bytes, bytearray]:  # type: ignore
                    adapters.append(lambda x, adapt=arg_type: adapt(x, "utf-8"))
                else:
                    adapters.append(lambda x, adapt=arg_type: adapt(x))

        return adapters, validators, arg_types, names

    def updates(
        self, response: Optional[GoProResp] = None
    ) -> Generator[tuple[enum.IntEnum, Any, GoProModel.Update], None, None]:
        """Generate updates from a response or any asynchronous updates

        Args:
            response (Optional[GoProResp], optional): Response to analyze. If none, get all asynchronous
                updates. Defaults to None.

        Yields:
            Generator[tuple[enum.IntEnum, Any, GoProModel.Update], None, None]: generates (updates identifier,
                update value, update type)
        """

        def get_update_type(container: GoProResp, identifier: ResponseType) -> GoProModel.Update:
            if container.protocol is GoProResp.Protocol.BLE:
                if container.cmd in [
                    constants.QueryCmdId.GET_CAPABILITIES_VAL,
                    constants.QueryCmdId.REG_CAPABILITIES_UPDATE,
                    constants.QueryCmdId.SETTING_CAPABILITY_PUSH,
                ]:
                    return GoProModel.Update.CAPABILITY
                if container.cmd in [
                    constants.QueryCmdId.GET_SETTING_VAL,
                    constants.QueryCmdId.REG_SETTING_VAL_UPDATE,
                    constants.QueryCmdId.SETTING_VAL_PUSH,
                ]:
                    return GoProModel.Update.SETTING
                if container.cmd in [
                    constants.QueryCmdId.GET_STATUS_VAL,
                    constants.QueryCmdId.REG_STATUS_VAL_UPDATE,
                    constants.QueryCmdId.STATUS_VAL_PUSH,
                ]:
                    return GoProModel.Update.STATUS
                # Must be protobuf
                return GoProModel.Update.PROTOBUF
            if container.protocol is GoProResp.Protocol.WIFI:
                if isinstance(identifier, constants.StatusId):
                    return GoProModel.Update.STATUS
                if isinstance(identifier, constants.SettingId):
                    return GoProModel.Update.SETTING
                raise TypeError(f"Received unexpected WiFi identifier: {identifier}")
            raise TypeError(f"Received unexpected protocol {container.protocol}")

        if isinstance(response, GoProResp):
            for identifier, value in response.items():
                if type(identifier) in (constants.SettingId, constants.StatusId):
                    yield identifier, value, get_update_type(response, identifier)
        elif response is None:
            while update := self.gopro.get_notification(0):
                for identifier, value in update.items():
                    yield identifier, value, get_update_type(update, identifier)


class CompoundCommand(Command):
    """Functionality that consists of multiple BLE and / or Wifi commands"""

    def __str__(self) -> str:
        return self._identifier

    def _as_dict(self, *_: Any, **kwargs: Any) -> dict[str, Any]:
        """Return the command as a dict

        Args:
            *_ (Any): unused
            **kwargs (Any) : additional dict keys to append

        Returns:
            dict[str, Any]: command as dict
        """
        return dict(protocol="Complex", id=self._identifier) | kwargs


# pylint: disable = missing-class-docstring
class CompoundCommands(Commands):
    """The container for the compound commands"""

    def __init__(self, communicator: GoPro) -> None:
        """Constructor

        Args:
            communicator (GoPro): the communicator to send the commands
        """

        class LiveStream(CompoundCommand):
            def __call__(
                self,
                ssid: str,
                password: str,
                url: str,
                window_size: Params.WindowSize,
                lens_type: Params.LensType,
                min_bit: int,
                max_bit: int,
                start_bit: int,
            ) -> GoProResp:
                """Disable shutter, connect to WiFi, start livestream, and set shutter

                Args:
                    ssid (str): SSID to connect to
                    password (str): password of WiFi network
                    url (str): url used to stream. Set to empty string to invalidate/cancel stream
                    window_size (open_gopro.api.params.WindowSize): Streaming video resolution
                    lens_type (open_gopro.api.params.LensType): Streaming Field of View
                    min_bit (int): Desired minimum streaming bitrate (>= 800)
                    max_bit (int): Desired maximum streaming bitrate (<= 8000)
                    start_bit (int): Initial streaming bitrate (honored if 800 <= value <= 8000)

                Returns:
                    GoProResp: status and url to start livestream
                """
                self._communicator.ble_command.set_shutter(Params.Toggle.DISABLE)
                self._communicator.ble_command.register_livestream_status([Params.RegisterLiveStream.STATUS])

                self._communicator.ble_command.scan_wifi_networks()
                # Wait to receive scanning success
                scan_id: Optional[int] = None
                while update := self._communicator.get_notification():
                    if (
                        update == constants.ActionId.NOTIF_START_SCAN
                        and update["scanning_state"] == Params.ScanState.SUCCESS
                    ):
                        scan_id = update["scan_id"]
                        break

                # Get scan results and see if we need to provision
                assert scan_id
                for entry in self._communicator.ble_command.get_ap_entries(scan_id)["entries"]:
                    if entry["ssid"] == ssid:
                        if not entry["scan_entry_flags"] & Params.ScanEntry.CONFIGURED:
                            self._communicator.ble_command.request_wifi_connect(ssid, password)
                            # Wait to receive provisioning done notification
                            while update := self._communicator.get_notification():
                                if (
                                    update == constants.ActionId.NOTIF_PROVIS_STATE
                                    and update["provisioning_state"] == Params.ProvisioningState.SUCCESS_NEW_AP
                                ):
                                    break
                        break

                # Start livestream
                self._communicator.ble_command.set_livestream_mode(
                    url=url,
                    window_size=window_size,
                    cert=bytes(),
                    minimum_bitrate=min_bit,
                    maximum_bitrate=max_bit,
                    starting_bitrate=start_bit,
                    lens=lens_type,
                )
                # Wait to receive livestream started status
                while update := self._communicator.get_notification():
                    if (
                        update == constants.ActionId.LIVESTREAM_STATUS_NOTIF
                        and update["live_stream_status"] == Params.LiveStreamStatus.READY
                    ):
                        break
                assert self._communicator.ble_command.set_shutter(Params.Toggle.ENABLE).is_ok

                response = GoProResp(meta=["Livestream"], raw_packet=dict(url=url))
                response._parse()
                return response

        self.livestream = LiveStream(communicator, "Livestream")

        super().__init__(communicator)
